/**
        \file arguments.c
        \author Sachith Dunatunga
        \date 03.04.09
        \version 0.2

        The implementation for command line parsing functions.
*/
#include <stdlib.h>
#include <string.h>
#include "arguments.h"

/* ---args----------------------------------------------------------------- */
/**
        \brief Gets the number of arguments in the line.
        \param line The line to get the number of arguments.
        \param fs The field separator to use.
        \param esc The escape character to use.
        \return 0 on error, number of arguments otherwise.
        \warning This function is for internal implementation only.

        This function will return the number of arguments that line contains;
        if line is invalid (unmatched quotes) or NULL, this function will
        return 0.
*/
int args_number(char *line, char fs, char esc);

int args_number(char *line, char fs, char esc)
{
        int len;
        int i;
        int inside_quote;
        int argc = 0;

        if (0 == line) {
                return argc;
        }

        len = strlen(line);

        /*
                Keep track of if we are inside quotes or not. Count the number
                of arguments while doing this.
        */
        for (i = 0, inside_quote = 0; i < len; i++) {
                if (esc == line[i]) {
                        /*
                                Since the next character is to be escaped, it
                                cannot affect the state; skip it.
                        */
                        i++;
                } else if ('\"' == line[i]) {
                        /*
                                If we weren't inside a quote, we are now;
                                conversely we are brought out of a quote if we
                                were in one.
                        */
                        inside_quote = !inside_quote;
                } else if (!inside_quote && i > 1 && fs == line[i]) {
                        /*
                                We are not inside a quote, so this is another
                                argument if the preceeding character is not
                                the field separator. Otherwise "   " would be
                                several arguments of size 0 (bad news).
                        */
                        if (fs != line[i - 1]) {
                                argc++;
                        }
                }
        }

        /*
                This method does not catch the last argument if it is not
                terminated by a field separator.
        */
        if (fs != line[len - 1]) {
                argc++;
        }

        /* There are unmatched quotes, so the line is invalid. */
        if (inside_quote) {
                argc = 0;
        }

        return argc;
}

args_t *args_parse(char *line, char fs, char esc)
{
        args_t *ret = NULL;
        char *tmp;
        int len, tmplen, inside_quote, escape;
        int alloc_failed = 0;
        int i, p, n;

        if (NULL == line) {
                return ret;
        }

        len = strlen(line);

        if (len <= 0) {
                return ret;
        }

        tmp = malloc(len + 1);
        if (NULL == tmp) {
                alloc_failed = 1;
                goto tmp_failed;
        }
        ret = malloc(sizeof(args_t));
        if (NULL == ret) {
                alloc_failed = 1;
                goto ret_failed;
        }
        ret->argc = args_number(line, fs, esc);
        if (ret->argc > 0) {
                ret->argv = calloc(ret->argc, sizeof(char *));
        } else {
                ret->argv = NULL;
        }
        if (NULL == ret->argv) {
                alloc_failed = 1;
                goto argv_failed;
        }

        /* Loop through and extract all the arguments. */
        inside_quote = escape = 0;
        n = p = 0;
        for (i = 0; i < len; i++) {
                if (escape) {
                        tmp[p++] = line[i];
                        escape = 0;
                } else if (esc == line[i]) {
                        /* Copy the next character blindly. */
                        escape = 1;
                } else if ('\"' == line[i]) {
                        /*
                                Keep track of the quotes so we don't
                                accidentally split arguments in the wrong
                                place.
                        */
                        inside_quote = !inside_quote;
                } else if (!inside_quote && i > 1 && fs == line[i]) {
                        /*
                                We are not inside a quote, so this is the end
                                of the argument.
                        */
                        tmp[p] = '\0';
                        tmplen = strlen(tmp);
                        if (tmplen > 0) {
                                ret->argv[n] = malloc(tmplen + 1);
                                if (NULL == ret->argv[n]) {
                                        alloc_failed = 1;
                                        goto argv_n_failed;
                                }
                                strncpy(ret->argv[n], tmp, tmplen + 1);
                                n++;
                        }
                        p = 0;
                } else {
                        tmp[p++] = line[i];
                }
        }

        /*
                This method does not catch the last argument if it is not
                terminated by a field separator, so we add it in now.
        */
        if (fs != line[len - 1]) {
                tmp[p] = '\0';
                tmplen = strlen(tmp);
                if (tmplen > 0) {
                        ret->argv[n] = malloc(tmplen + 1);
                        if (NULL == ret->argv[n]) {
                                alloc_failed = 1;
                                goto argv_n_failed;
                        }
                        strncpy(ret->argv[n], tmp, tmplen + 1);
                }
        }

        free(tmp);

        /*
                We can't get here normally, but we don't want to call this
                code unless something goes wrong and we can't allocate memory.
        */
        if (alloc_failed) {
                argv_n_failed:
                i = n;
                while (i > 0) {
                        free(ret->argv[--i]);
                }
                free(ret->argv);
                argv_failed:
                free(ret);
                ret_failed:
                free(tmp);
                tmp_failed:
                ret = NULL;
        }

        return ret;
}

args_t *args_parse_simple(char *line)
{
        return args_parse(line, ' ', '\\');
}

void args_cleanup(args_t *args)
{
        int i;

        if (NULL == args) {
                return;
        }

        if (NULL == args->argv) {
                free(args);
                return;
        }

        for (i = 0; i < args->argc; i++) {
                free(args->argv[i]);
        }
        free(args->argv);
        free(args);

        return;
}
/* ------------------------------------------------------------------------ */
