/**
        \file arguments.h
        \author Sachith Dunatunga
        \date 03.04.09
        \version 1.0

        The header/interface for command line parsing functions.
*/
#ifndef __ARGUMENTS_H__
#define __ARGUMENTS_H__
/* ---args----------------------------------------------------------------- */
/**
        \brief A structure to hold command line arguments.

        This structure tries to keep the familiar argv/argc argument
        convention intact.
*/
struct args_s {
        /**
                \brief The array of arguments.

                All of the arguments in the array are null terminated,
                although the argument list itself is not null terminated.
        */
        char **argv;
        /**
                \brief The number of arguments.
        */
        int argc;
};

/**
        \brief This is a typedef for the arguments structure. It is used for
        clarity in the program implementation.
*/
typedef struct args_s args_t;

/**
        \brief Parses a single line into the args_t format.
        \param line The line to parse. It should not include the newline.
        \param fs The field separator -- determines where to split arguments.
        \param esc The escape character -- allows quotes and fs characters to
        be part of arguments.
        \return NULL on error, a args_t pointer otherwise.

        The parse function simply scans the line looking for the field
        separator, quotes, and the escape character. It keeps track of the
        state and decides where the arguments should be split off, and returns
        the structure for further processing by another function. An important
        feature to note about this parsing function is that only double quotes
        can 'quote' something; that is, single quotes are treated like any
        other character and we cannot nest quotes to get multiple levels.
        Also, the parsing function strips the actual quote character from the
        argument, so if your line contains: \code "arg1" \endcode the
        corresponding argument will be: \code arg1 \endcode If you need the
        quotes, you should escape them first, as with: \code \"arg1\" \endcode
        which produces: \code "arg1" \endcode
*/
args_t *args_parse(char *line, char fs, char esc);

/**
        \brief Parses a line into the args_t format using default values
        for fs and esc.
        \param line The line to parse. It should not include the newline.
        \return NULL on error, a args_t pointer otherwise.

        This function just calls the args_parse() function with a space for
        the field separator and a backslash for the escape character. As a
        result it has all the quirks of the args_parse() function.
*/
args_t *args_parse_simple(char *line);

/**
        \brief Frees all the memory associated with the args_t structure.
        \param args The args_t structure to free.

        This function deallocates memory for the arguments and the argument
        array.
*/
void args_cleanup(args_t *args);
/* ------------------------------------------------------------------------ */
#endif
