/**
        \file command.c
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The implementation of the paragon commands to send to the
        phone.
*/
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "command.h"
#include "error.h"
#include "logger.h"
#include "handle.h"
#include "modes.h"
#include "timings.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief This function is a helper for paragon_send_command() for p2k
        phones.
        \param handle The libparagon handle to use.
        \param cmd The command issued.
        \param recv A pointer to an unsigned character array. Used to hold
        the response.
        \param recv_size A pointer to put the number of bytes in '*recv'.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        This function is used to create and send the binary packet to p2k
        phones.
*/
int paragon_send_p2k(paragon_handle_t *handle, paragon_command_t *cmd,
        unsigned char **recv, int *recv_size);
/**
        \brief This function is a helper for paragon_send_command() for p2k05
        phones.
        \param handle The libparagon handle to use.
        \param cmd The command issued.
        \param recv A pointer to an unsigned character array. Used to hold
        the response.
        \param recv_size A pointer to put the number of bytes in '*recv'.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        This function is used to create and send the binary packet to p2k05
        phones.
*/
int paragon_send_p2k05(paragon_handle_t *handle, paragon_command_t *cmd,
        unsigned char **recv, int *recv_size);

/**
        \brief The header size (in bytes) for Paragon 2005 phones.
*/
const int PARAGON_2005_HEADER_SIZE = 12;

int paragon_send_p2k(paragon_handle_t *handle, paragon_command_t *cmd,
        unsigned char **recv, int *recv_size)
{
        int result = PARAGON_OPENUSB_ERROR;

        assert(NULL != handle);
        assert(NULL != cmd);
        assert(NULL != recv);
        assert(NULL != recv_size);

        return result;
}

int paragon_send_p2k05(paragon_handle_t *handle, paragon_command_t *cmd,
        unsigned char **recv, int *recv_size)
{
        struct openusb_request_handle rqhandle;
        openusb_bulk_request_t bulk;
        int result = PARAGON_OPENUSB_ERROR;
        unsigned char *buffer;
        unsigned char *response;
        int buflen;
        int responselen;

        assert(NULL != handle);
        assert(NULL != cmd);
        assert(NULL != recv);
        assert(NULL != recv_size);

        /*
                Here we calculate the size of the buffer we send to the phone.
                The format of the command is:
                Byte 0: Status Byte
                        Bit 0: Unsolicted Response (1 = Yes; 0 = No)
                        Bit 1: Data Follows (1 = Yes; 0 = No)
                        Bit 2: Command Failed (1 = Yes; 0 = No)
                        Bits 3...6: Test Command Set Version
                        Bit 7: Direction (1 = To Computer; 0 = To Phone)
                Byte 1: PacketID
                Bytes 2, 3: Command
                Bytes 4...7: Zero Padding (0)
                Bytes 8...11: Argument Size (Bytes)
                Bytes 12...: Arguments (Data)

                Note that all multibyte entities are big endian.

                If we have an FSAC command, the format is:
                Byte 0: Status Byte
                        Bit 0: Unsolicted Response (1 = Yes; 0 = No)
                        Bit 1: Data Follows (1 = Yes; 0 = No)
                        Bit 2: Command Failed (1 = Yes; 0 = No)
                        Bits 3...6: Test Command Set Version
                        Bit 7: Direction (1 = To Computer; 0 = To Phone)
                Byte 1: PacketID
                Bytes 2, 3: Command (0x4A)
                Bytes 4...7: Zero Padding (0)
                Bytes 8...11: Argument Size (Bytes)
                Bytes 12, 13, 14, 15: FSAC Command
                Bytes 16...: Arguments (Data)
                Note that all multibyte entities are big endian.
        */

        if (cmd->command != PARAGON_COMMAND_FSAC) {
                buflen = 12 + cmd->extra_len;
        } else {
                buflen = 16 + cmd->extra_bytes.fsac.data_len;
        }
        buffer = malloc(buflen);

        if (NULL == buffer) {
                return PARAGON_NO_MEMORY;
        }

        /*
                Unfortunately with Paragon 2005 phones I don't know what
                the command is to figure out how much data is waiting. This
                will frequently be much too large.
        */
        response = malloc(PARAGON_MAX_PACKET_SIZE + PARAGON_2005_HEADER_SIZE);

        if (NULL == response) {
                free(buffer);
                return PARAGON_NO_MEMORY;
        }

        /*
                Here we setup the packet according to structure defined
                before.
        */
        buffer[0] = 0x00;
        buffer[1] = handle->packet_id;

        BE_SET_INT_16(buffer, 2, cmd->command);
        BE_SET_INT_32(buffer, 4, 0);

        if (cmd->command != PARAGON_COMMAND_FSAC) {
                BE_SET_INT_32(buffer, 8, cmd->extra_len);
                if (NULL != cmd->extra_bytes.data && cmd->extra_len > 0) {
                        memcpy(buffer + 12, cmd->extra_bytes.data,
                                cmd->extra_len);
                }
        } else {
                /*
                        The FSAC command is a 32-bit integer, so account for
                        it in the argument size.
                */
                BE_SET_INT_32(buffer, 8, 4 + cmd->extra_bytes.fsac.data_len);
                BE_SET_INT_32(buffer, 12, cmd->extra_bytes.fsac.command);
                if (NULL != cmd->extra_bytes.fsac.data &&
                        cmd->extra_bytes.fsac.data_len > 0) {
                        memcpy(buffer + 16, cmd->extra_bytes.fsac.data,
                                cmd->extra_bytes.fsac.data_len);
                }
        }

        /*
                Increment our packet ID for the next packet. Make sure it's
                in the correct range and doesn't overflow.
        */
        handle->packet_id++;
        handle->packet_id &= 0xFF;

        paragon_log(handle, PL_INFO, "Sending buffer:\n");
        paragon_print_hexdump(handle, buffer, buflen);

        /*
                Setup our bulk transfer to the phone.
        */
        memset(&bulk, 0, sizeof(bulk));
        bulk.payload = buffer;
        bulk.timeout = PARAGON_BULK_DEFAULT_TIMEOUT;
        bulk.length = buflen;

        /*
                Send it to the right interface and the host to device
                endpoint.
        */
        rqhandle.dev = handle->hdev;
        rqhandle.interface = handle->ifnum;
        rqhandle.endpoint = handle->ep_out;
        rqhandle.type = USB_TYPE_BULK;
        rqhandle.req.bulk = &bulk;

        /*
                Write the data to the phone; bail if it doesn't work.
        */
        result = openusb_xfer_wait(&rqhandle);
        free(buffer);

        if (OPENUSB_SUCCESS != result) {
                paragon_log(handle, PL_ERROR, "Sending bulk transfer failed:"
                        " %s\n", openusb_strerror(result));
                return PARAGON_OPENUSB_ERROR;
        } else {
                result = PARAGON_SUCCESS;
        }

        memset(&bulk, 0, sizeof(bulk));
        bulk.payload = response;
        bulk.timeout = PARAGON_BULK_DEFAULT_TIMEOUT;
        bulk.length = PARAGON_MAX_PACKET_SIZE + PARAGON_2005_HEADER_SIZE;

        /*
                Send it to the right interface and the device to host
                endpoint.
        */
        rqhandle.dev = handle->hdev;
        rqhandle.interface = handle->ifnum;
        rqhandle.endpoint = handle->ep_in;
        rqhandle.type = USB_TYPE_BULK;
        rqhandle.req.bulk = &bulk;

        /*
                Read data from the phone; bail if it doesn't work.
        */
        result = openusb_xfer_wait(&rqhandle);
        if (OPENUSB_SUCCESS != result) {
                paragon_log(handle, PL_ERROR, "Reading bulk transfer failed:"
                        " %s\n", openusb_strerror(result));
                return PARAGON_OPENUSB_ERROR;
        } else {
                result = PARAGON_SUCCESS;
        }

        responselen = bulk.result.transferred_bytes;

        paragon_log(handle, PL_INFO, "Received buffer (Full Packet):\n");
        paragon_print_hexdump(handle, response, responselen);

        /*
                There is a 12 byte response header on Paragon 2005 phones. I
                don't know the exact construction -- from inspection it
                seems:

                Byte 0: Status Byte
                        Bit 0: Unsolicted Response (1 = Yes; 0 = No)
                        Bit 1: Data Follows (1 = Yes; 0 = No)
                        Bit 2: Command Failed (1 = Yes; 0 = No)
                        Bits 3...6: Test Command Set Version (Normally 0000)
                        Bit 7: Direction (1 = To Computer; 0 = To Phone)
                Byte 1: PacketID (most recently sent packet id)
                Bytes 2, 3: Command (most recently sent command)
                Byte 4: Zero Padding (0)
                Byte 5: Response Code (06 Generic Success, 07 Generic Failure)
                Bytes 6, 7: Zero Padding (0)
                Bytes 8...11: Data length (data follows this byte)

                The multibyte values are in big endian format.
        */
        if (response[0] & COMMAND_FAILED || response[5] == GENERIC_FAILURE) {
                paragon_log(handle, PL_WARNING, "Command (0x%04x) failed"
                        " with status (0x%02x) and response code (0x%02x).\n",
                        BE_GET_INT_16(response, 2), response[0],
                        response[5]);
                free(response);
                return PARAGON_COMMAND_ERROR;
        }

        if (response[0] & DATA_FOLLOWS) {
                *recv_size = BE_GET_INT_32(response, 8);
        } else {
                *recv_size = 0;
                result = PARAGON_SUCCESS;
        }

        if (*recv_size > 0) {
                *recv = malloc(*recv_size);
                if (NULL == *recv) {
                        free(response);
                        return PARAGON_NO_MEMORY;
                }

                memcpy(*recv, response + PARAGON_2005_HEADER_SIZE,
                        *recv_size);
                result = PARAGON_SUCCESS;

                paragon_log(handle, PL_INFO,
                        "Received buffer (Data section):\n");
                paragon_print_hexdump(handle, *recv, *recv_size);
        } else {
                *recv = NULL;
        }

        free(response);
        return result;
}

int paragon_send_command(paragon_handle_t *handle, paragon_command_t *cmd,
        unsigned char **recv, int *recv_size)
{
        int result = PARAGON_NO_PHONE;
        int mode = PARAGON_MODE_UNKNOWN;

        /*
                We need a good handle for this and actual data.
        */
        if (NULL == handle || NULL == cmd || NULL == recv ||
                NULL == recv_size) {
                return PARAGON_BADARG;
        }

        result = paragon_get_mode(handle, &mode);

        if (PARAGON_SUCCESS != result) {
                return result;
        }

        if (PARAGON_MODE_P2K != mode) {
                return PARAGON_MODE_ERROR;
        }

        /*
                On Paragon 2000 phones, the address of the 'in' endpoint
                of the Test Command device is 0x00; it is another value on
                Paragon 2005 phones (on mine it is 0x01);
        */
        if (0x00 == handle->ep_in) {
                result = paragon_send_p2k(handle, cmd, recv, recv_size);
        } else {
                result = paragon_send_p2k05(handle, cmd, recv, recv_size);
        }

        return result;
}
/* ------------------------------------------------------------------------ */
