/**
        \file command.h
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The header/interface for paragon commands to send to the
        phone.
*/
#ifndef __PARAGON_COMMAND_H__
#define __PARAGON_COMMAND_H__
#include "handle.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief This structure is used to send FSAC commands to the phone.

        We use this structure extensively for file system access comamnds.
*/
struct paragon_fsac_s {
        /**
                \brief The FSAC command to issue.
        */
        int command;
        /**
                \brief The data for the FSAC command.
        */
        unsigned char *data;
        /**
                \brief The length (in bytes) of the data.
        */
        int data_len;
};

/**
        \brief This structure is used to send different commands and data
        to the phone.

        The paragon_send_command() function uses data in this structure to
        generate and send a phone-specific binary packet.
*/
struct paragon_command_s {
        /**
                \brief The command code to issue.

                This is the P2K/P2K05 command to issue. We only know about a
                handful of commands at the moment. If this is the FSAC
                command, we will use extra_bytes.fsac to determine other
                parameters; otherwise, we will simply copy 'data' into the
                binary packet.
        */
        int command;
        /**
                \brief This union contains either extra data or the FSAC
                command to issue.
        */
        union {
                /**
                        \brief The parameters for the FSAC command we issue.
                */
                struct paragon_fsac_s fsac;
                /**
                        \brief The data for this command.
                */
                unsigned char *data;
        } extra_bytes;
        /**
                \brief The length of 'extra_bytes.data'.

                If this is not an FSAC command, we use this to determine how
                many bytes to copy from 'extra_bytes.data'.
        */
        int extra_len;
};

/**
        \brief A typedef to improve readability.
*/
typedef struct paragon_command_s paragon_command_t;
/**
        \brief A typedef to improve readability.
*/
typedef struct paragon_fsac_s paragon_fsac_t;

/**
        \defgroup p2k_commands P2K/P2K05 Commands
                The list of possible commands to issue to the phone.
        @{
*/
/**
        \brief This command is used to request a seem read.
*/
#define PARAGON_COMMAND_READ_SEEM (0x20)
/**
        \brief This command is used to request a seem write.
*/
#define PARAGON_COMMAND_WRITE_SEEM (0x2F)
/**
        \brief This command is used to send an FSAC command.
*/
#define PARAGON_COMMAND_FSAC (0x4A)
/**
        @}
*/

/**
        \brief The maximum size data packet returned by a Paragon 2000 /
        Paragon 2005 Phone.
*/
#define PARAGON_MAX_PACKET_SIZE 0x400

/**
        \brief Big Endian Get 16-bit integer. Used for getting data from
        the binary format packets.
*/
#define BE_GET_INT_16(buf, p) ((buf[(p)] << 8) | (buf[(p) + 1]))
/**
        \brief Big Endian Set 16-bit integer. Used for setting data for
        the binary format packets.
*/
#define BE_SET_INT_16(buf, p, v) ({ \
        buf[(p)] = ((v) >> 8) & 0xFF; \
        buf[(p) + 1] = (v) & 0xFF; \
})
/**
        \brief Big Endian Get 32-bit integer. Used for getting data from
        the binary format packets.
*/
#define BE_GET_INT_32(buf, p) ((buf[(p)] << 24) | (buf[(p) + 1] << 16 ) | \
        (buf[(p) + 2] << 8) | (buf[(p) + 3]))
/**
        \brief Big Endian Set 32-bit integer. Used for setting data for
        the binary format packets.
*/
#define BE_SET_INT_32(buf, p, v) ({ \
        buf[(p)] = ((v) >> 24) & 0xFF; \
        buf[(p) + 1] = ((v) >> 16) & 0xFF; \
        buf[(p) + 2] = ((v) >> 8) & 0xFF; \
        buf[(p) + 3] = (v) & 0xFF; \
})

/**
        \brief If this bit is set in the status byte, data follows the
        header.
*/
#define DATA_FOLLOWS (0x02)
/**
        \brief If this bit is set in the status byte, the command failed.
*/
#define COMMAND_FAILED (0x04)
/**
        \brief If this is the response code, the operation was successful.
*/
#define GENERIC_SUCCESS (0x06)
/**
        \brief If this is the response code, the operation failed.
*/
#define GENERIC_FAILURE (0x07)

/**
        \brief This function sends a command to the phone.
        \param handle The libparagon handle to use.
        \param cmd The command to send to the phone.
        \param recv A pointer to an unsigned character array. Used to hold
        the response.
        \param recv_size A pointer to put the number of bytes in '*recv'.
        \return PARAGON_SUCCESS on success, an integer less than 0 on
        failure.
        \warning Don't forget to call free on '*recv' when you are done with
        it.

        This function assumes that the phone pointed to by 'handle' is in
        P2K mode; if it is not, this function will return PARAGON_MODE_ERROR.
        It builds a raw packet from the command structure and sends it to
        the correct endpoint/interface. It automatically determines the
        correct binary format for the phone (which changed between Paragon
        2000 and Paragon 2005 phones). The received response is just the data
        portion of the full packet received; thus, it is independent of the
        underlying phone type.
*/
int paragon_send_command(paragon_handle_t *handle, paragon_command_t *cmd,
        unsigned char **recv, int *recv_size);
/* ------------------------------------------------------------------------ */
#endif
