/**
        \file fsac.c
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The implementation of the libparagon commands to manipulate
        files and directories.
*/
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "command.h"
#include "error.h"
#include "fsac.h"
#include "logger.h"
#include "handle.h"
#include "seem.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief The number of bytes needed in the data section of the get
        volumes command. Think of them as the argument bytes.
*/
const int PARAGON_GET_VOLUMES_DATA_LEN = 0;
/**
        \brief The number of bytes needed in the data section of the get
        free space command. Think of them as the argument bytes.
*/
const int PARAGON_GET_FREE_SPACE_DATA_LEN = 8;
/**
        \brief The mininum number of bytes needed in the data section of the
        get number of files command. Think of them as the argument bytes.
*/
const int PARAGON_GET_FILES_NUM_MIN_DATA_LEN = 10;
/**
        \brief The number of bytes needed in the data section of the get file
        record command. Think of them as the argument bytes.
*/
const int PARAGON_GET_FILE_RECORD_DATA_LEN = 1;
/**
        \brief The mininum number of bytes needed in the data section of the
        open file command. Think of them as the argument bytes.
*/
const int PARAGON_OPEN_FILE_MIN_DATA_LEN = 4;
/**
        \brief The number of bytes needed in the data section of the read file
        command. Think of them as the argument bytes.
*/
const int PARAGON_READ_FILE_DATA_LEN = 4;
/**
        \brief The minimum number of bytes needed in the data section of the
        write file command. Think of them as the argument bytes.
*/
const int PARAGON_WRITE_FILE_MIN_DATA_LEN = 4;
/**
        \brief The number of bytes needed in the data section of the seek file
        command. Think of them as the argument bytes.
*/
const int PARAGON_SEEK_FILE_DATA_LEN = 5;
/**
        \brief The mininum number of bytes needed in the data section of the
        delete file command. Think of them as the argument bytes.
*/
const int PARAGON_DELETE_FILE_MIN_DATA_LEN = 0;
/**
        \brief The mininum number of bytes needed in the data section of the
        create directory command. Think of them as the argument bytes.
*/
const int PARAGON_CREATE_DIR_MIN_DATA_LEN = 4;
/**
        \brief The mininum number of bytes needed in the data section of the
        delete directory command. Think of them as the argument bytes.
*/
const int PARAGON_DELETE_DIR_MIN_DATA_LEN = 4;

/**
        \brief The number of bytes returned in the data section of the get
        free space command.
*/
const int PARAGON_GET_FREE_SPACE_RETURN_DATA_LEN = 4;
/**
        \brief The number of bytes returned in the data section of the get
        number of files command.
*/
const int PARAGON_GET_FILES_NUM_RETURN_DATA_LEN = 2;
/**
        \brief The number of bytes returned in the data section of the get
        file record command.
*/
const int PARAGON_GET_FILE_RECORD_RETURN_DATA_LEN = 272;

/**
        \brief The maximum size of a single read command. This is a hard
        limit built into the phone.
*/
const int PARAGON_MAX_SINGLE_READ_SIZE = 0x400;
/**
        \brief The maximum size of a single write command. This is a hard
        limit built into the phone.
*/
const int PARAGON_MAX_SINGLE_WRITE_SIZE = 0x400;

/**
        \brief The UCS-2 version of the slash '/'.
*/
const int PARAGON_UCS2_SLASH = 0x002F;
/**
        \brief This sequence separates arguments in the data section. It is
        the Unicode Byte Order Marker.
*/
const int PARAGON_SEPARATOR = 0xFFFE;

int paragon_get_volumes(paragon_handle_t *handle, char **volumes,
        int *volumes_len)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;
        int i;

        if (NULL == handle || NULL == volumes || NULL == volumes_len) {
                return PARAGON_BADARG;
        }

        /*
                Interestingly enough, the 'get volumes' fsac command doesn't
                take any parameters. The paragon_send_command() function
                will correctly handle this case, however.
        */
        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_GET_VOLUMES;
        command.extra_bytes.fsac.data_len = 0;
        command.extra_bytes.fsac.data = NULL;

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                return result;
        }

        /*
                The returned buffer has the following structure:

                Bytes 0...3: 1st full volume path in UCS-2
                Byte 4: 0xff
                Byte 5: 0xfe
                Bytes 6...9: 2nd full volume path in UCS-2
                Byte 10: 0xff
                Byte 11: 0xfe
                Bytes 12...15: 3rd full volume path in UCS-2
                Bytes 16...n-2: [continues above pattern]
                Bytes n-1, n: zero padding (0)

                The last full volume path is terminated by two zero bytes
                (corresponding to the null terminator in UCS-2) instead of
                0xff 0xfe.

                We can quickly convert UCS-2 to ascii by stripping out every
                even byte if it's zero (if it's not zero, it doesn't exist
                in the ascii set and we should do something to signal that).

                Thus, starting from n = 3, every 6th byte is another
                volume name (without root directory).
        */

        /*
                Calculate the number of volumes based on the number of bytes
                we got back. From inspection of the structure above, we can
                see this formula yields the number of volumes.
        */
        *volumes_len = ((buflen) / 6);
        if (*volumes_len < 1) {
                free(buffer);
                return PARAGON_FSAC_ERROR;
        }

        *volumes = malloc(*volumes_len);
        if (NULL == *volumes) {
                free(buffer);
                return PARAGON_NO_MEMORY;
        }

        /*
                This is a very crude way to extract all the volumes from the
                buffer. Extract every 6th byte starting from the 3rd byte
                (see the structure of the buffer for explanation).
        */
        for (i = 0; i < *volumes_len; i++) {
                (*volumes)[i] = buffer[3 + (i * 6)];
        }

        free(buffer);
        return result;
}

int paragon_get_free_space(paragon_handle_t *handle, char volume,
        int *freespace)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle || NULL == freespace) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_GET_FREE_SPACE;
        command.extra_bytes.fsac.data_len = PARAGON_GET_FREE_SPACE_DATA_LEN;
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Bytes 0, 1: zero padding (0)
                Bytes 2, 3: UCS-2 version of '/' (00 2F)
                Bytes 4, 5: UCS-2 version of the volume character (00 xx),
                        where xx is the value corresponding to the ascii
                        encoding of 'volume'
                Bytes 6, 7: zero padding (0)
        */
        BE_SET_INT_16(command.extra_bytes.fsac.data, 0, 0x0000);
        BE_SET_INT_16(command.extra_bytes.fsac.data, 2, PARAGON_UCS2_SLASH);
        command.extra_bytes.fsac.data[4] = 0x00;
        command.extra_bytes.fsac.data[5] = volume;
        BE_SET_INT_16(command.extra_bytes.fsac.data, 6, 0x0000);

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                The returned buffer has the following structure:

                Bytes 0...3: free space in bytes (32 bit integer)
        */
        if (PARAGON_GET_FREE_SPACE_RETURN_DATA_LEN != buflen) {
                free(buffer);
                free(command.extra_bytes.fsac.data);
                return PARAGON_FSAC_ERROR;
        }

        *freespace = BE_GET_INT_32(buffer, 0);

        free(buffer);
        free(command.extra_bytes.fsac.data);
        return result;
}

int paragon_set_file_mask(paragon_handle_t *handle, char volume, char *mask,
        int *filecount)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;
        int masklen;
        int i;

        if (NULL == handle || NULL == mask || NULL == filecount) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_GET_FILES_NUM;
        /*
                We need to convert the mask to UCS-2, so we need twice as
                many bytes.
        */
        command.extra_bytes.fsac.data_len =
                PARAGON_GET_FILES_NUM_MIN_DATA_LEN + 2 * strlen(mask);
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Bytes 0, 1: UCS-2 version of '/' (00 2F)
                Bytes 2, 3: UCS-2 version of the volume character (00 xx),
                        where xx is the value corresponding to the ascii
                        encoding of 'volume'
                Bytes 4, 5: UCS-2 version of '/' (00 2F)
                Bytes 6, 7: separator (FF FE)
                Bytes 8...n-2: UCS-2 version of 'mask' (usually "*")
                Bytes n-1, n: zero padding (0)
        */
        BE_SET_INT_16(command.extra_bytes.fsac.data, 0, PARAGON_UCS2_SLASH);
        command.extra_bytes.fsac.data[2] = 0x00;
        command.extra_bytes.fsac.data[3] = volume;
        BE_SET_INT_16(command.extra_bytes.fsac.data, 4, PARAGON_UCS2_SLASH);
        BE_SET_INT_16(command.extra_bytes.fsac.data, 6, PARAGON_SEPARATOR);
        /*
                Here is a crude way to convert the filter to UCS-2 and set
                it in the right location in the buffer.
        */
        masklen = strlen(mask);
        for (i = 0; i < masklen; i++) {
                command.extra_bytes.fsac.data[8 + (2 * i)] = 0x00;
                command.extra_bytes.fsac.data[9 + (2 * i)] = mask[i];
        }
        BE_SET_INT_16(command.extra_bytes.fsac.data, 8 + (2 * i), 0x00);

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                The returned buffer has the following structure:

                Bytes 0, 1: file count (16 bit integer)
        */
        if (PARAGON_GET_FILES_NUM_RETURN_DATA_LEN != buflen) {
                free(buffer);
                free(command.extra_bytes.fsac.data);
                return PARAGON_FSAC_ERROR;
        }

        *filecount = BE_GET_INT_16(buffer, 0);

        free(buffer);
        free(command.extra_bytes.fsac.data);
        return result;
}

int paragon_get_next_file_record(paragon_handle_t *handle,
        paragon_file_record_t *record)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle || NULL == record) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_GET_FILE_RECORD;
        command.extra_bytes.fsac.data_len = PARAGON_GET_FILE_RECORD_DATA_LEN;
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Byte 0: how many file records to get (we just want one in
                this function)
        */
        command.extra_bytes.fsac.data[0] = 1;

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                The returned buffer has the following structure:

                Bytes 0, 1: number of records in buffer
                Bytes 2, 3: zero padding (0)
                Bytes 4...263: The absolute path.
                Bytes 264, 265: The attributes (similar to permissions).
                Bytes 266, 267: The owner.
                Bytes 268..,271: The size (in bytes).
        */
        if (PARAGON_GET_FILE_RECORD_RETURN_DATA_LEN != buflen) {
                free(buffer);
                free(command.extra_bytes.fsac.data);
                return PARAGON_FSAC_ERROR;
        }

        memcpy(&(record->path), buffer + 4, PARAGON_FSAC_MAX_PATH_LEN);
        record->attributes = BE_GET_INT_16(buffer, 264);
        record->owner = BE_GET_INT_16(buffer, 266);
        record->size= BE_GET_INT_32(buffer, 268);

        free(buffer);
        free(command.extra_bytes.fsac.data);
        return result;
}

int paragon_open_file(paragon_handle_t *handle, char *path, int attributes)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle || NULL == path) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_OPEN_FILE;
        command.extra_bytes.fsac.data_len =
                PARAGON_OPEN_FILE_MIN_DATA_LEN + strlen(path);
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Bytes 0...3: The attributes (a 32-bit integer).
                Bytes 4...n: The absolute path to the file.
        */
        BE_SET_INT_32(command.extra_bytes.fsac.data, 0, attributes);
        memcpy(command.extra_bytes.fsac.data + 4, path, strlen(path));

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                This command usually does not return a data buffer, but can
                return something if there is an error. Unfortunately I don't
                know how to interpret the error, so it's pretty useless.
        */
        if (NULL != buffer) {
                free(buffer);
        }

        free(command.extra_bytes.fsac.data);
        return result;
}

int paragon_close_file(paragon_handle_t *handle)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle) {
                return PARAGON_BADARG;
        }

        /*
                Interestingly enough, the 'close file' fsac command doesn't
                take any parameters. The paragon_send_command() function
                will correctly handle this case, however.
        */
        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_CLOSE_FILE;
        command.extra_bytes.fsac.data_len = 0;
        command.extra_bytes.fsac.data = NULL;

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                return result;
        }

        /*
                We don't care about the returned buffer in this case. Indeed,
                it will only return a buffer if there is an error. I don't
                know how to parse the returned error though.
        */

        if (NULL != buffer) {
                free(buffer);
        }

        return result;
}

int paragon_read_file(paragon_handle_t *handle, int bytes,
        unsigned char **recv, int *recv_size)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;
        int num_full_reads;
        int partial_read_bytes;
        int bytes_read;
        int i;

        if (NULL == handle || NULL == recv || NULL == recv_size ||
                bytes < 1) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_READ_FILE;
        command.extra_bytes.fsac.data_len = PARAGON_READ_FILE_DATA_LEN;
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        *recv_size = bytes;
        /*
                Here we calculate how many runs of the basic function we need
                in order to get the requested number of bytes; the basic
                function can only transfer 1024 bytes at a time
                (a single packet).
        */
        num_full_reads = bytes / PARAGON_MAX_SINGLE_READ_SIZE;
        partial_read_bytes = bytes % PARAGON_MAX_SINGLE_READ_SIZE;
        *recv = malloc(*recv_size);

        if (NULL == *recv) {
                free(command.extra_bytes.fsac.data);
                return PARAGON_NO_MEMORY;
        }

         /*
                We set the data bytes according to this structure:

                Bytes 0...3: The number of bytes to read at once
                        (a 32-bit integer); the maximum value is 1024 bytes.
        */
        BE_SET_INT_32(command.extra_bytes.fsac.data, 0,
                PARAGON_MAX_SINGLE_READ_SIZE);

        bytes_read = 0;
        for (i = 0; i < num_full_reads; i++) {
                result = paragon_send_command(handle, &command, &buffer,
                        &buflen);

                if (PARAGON_SUCCESS != result) {
                        free(command.extra_bytes.fsac.data);
                        free(*recv);
                        *recv = NULL;
                        *recv_size = 0;
                        return result;
                }

                /*
                        All the returned bytes are simply file data; there is
                        no special formatting.
                */
                memcpy(*recv + bytes_read, buffer,
                        PARAGON_MAX_SINGLE_READ_SIZE);

                bytes_read += PARAGON_MAX_SINGLE_READ_SIZE;

                free(buffer);
        }

        if (partial_read_bytes > 0) {
                BE_SET_INT_32(command.extra_bytes.fsac.data, 0,
                        partial_read_bytes);

                result = paragon_send_command(handle, &command, &buffer,
                        &buflen);

                if (PARAGON_SUCCESS != result) {
                        free(command.extra_bytes.fsac.data);
                        free(*recv);
                        *recv = NULL;
                        *recv_size = 0;
                        return result;
                }

                memcpy(*recv + bytes_read, buffer, partial_read_bytes);
                free(buffer);
        }

        free(command.extra_bytes.fsac.data);

        paragon_log(handle, PL_INFO, "File Read Buffer:\n");
        paragon_print_hexdump(handle, *recv, *recv_size);

        return result;
}

int paragon_write_file(paragon_handle_t *handle, unsigned char *trans,
        int trans_size)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;
        int num_full_writes;
        int partial_write_bytes;
        int bytes_written;
        int i;

        if (NULL == handle || NULL == trans || 0 == trans_size) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_WRITE_FILE;
        command.extra_bytes.fsac.data_len = PARAGON_WRITE_FILE_MIN_DATA_LEN +
                PARAGON_MAX_SINGLE_WRITE_SIZE;
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                Here we calculate how many runs of the basic function we need
                in order to send the requested number of bytes; the basic
                function can only transfer 1024 bytes at a time
                (a single packet). Actually it can transfer 1024 + 12 bytes
                (for the header) but the header is mostly useless for our
                purposes.
        */
        num_full_writes = trans_size / PARAGON_MAX_SINGLE_WRITE_SIZE;
        partial_write_bytes = trans_size % PARAGON_MAX_SINGLE_WRITE_SIZE;
         /*
                We set the data bytes according to this structure:

                Bytes 0...3: The number of bytes in this packet.
                Bytes 4...n: The data to write to the file.
        */
        bytes_written = 0;
        BE_SET_INT_32(command.extra_bytes.fsac.data, 0,
                PARAGON_MAX_SINGLE_WRITE_SIZE);
        for (i = 0; i < num_full_writes; i++) {
                memcpy(command.extra_bytes.fsac.data +
                        PARAGON_WRITE_FILE_MIN_DATA_LEN,
                        trans + bytes_written,
                        PARAGON_MAX_SINGLE_WRITE_SIZE);
                result = paragon_send_command(handle, &command, &buffer,
                        &buflen);

                if (PARAGON_SUCCESS != result) {
                        free(command.extra_bytes.fsac.data);
                        return result;
                }

                bytes_written += PARAGON_MAX_SINGLE_WRITE_SIZE;

                /*
                        We don't care about the returned data.
                */
                free(buffer);
        }

        if (partial_write_bytes > 0) {
                /*
                        Be sure to alter the command structure to be the
                        correct number of bytes.
                */
                BE_SET_INT_32(command.extra_bytes.fsac.data, 0,
                        partial_write_bytes);
                command.extra_bytes.fsac.data_len =
                        PARAGON_WRITE_FILE_MIN_DATA_LEN +
                        partial_write_bytes;
                memcpy(command.extra_bytes.fsac.data +
                        PARAGON_WRITE_FILE_MIN_DATA_LEN,
                        trans + bytes_written,
                        partial_write_bytes);
                result = paragon_send_command(handle, &command, &buffer,
                        &buflen);

                if (PARAGON_SUCCESS != result) {
                        free(command.extra_bytes.fsac.data);
                        return result;
                }

                free(buffer);
        }

        free(command.extra_bytes.fsac.data);

        paragon_log(handle, PL_INFO, "File Write Buffer:\n");
        paragon_print_hexdump(handle, trans, trans_size);

        return result;
}

int paragon_seek_file(paragon_handle_t *handle, int offset, int direction)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_SEEK_FILE;
        command.extra_bytes.fsac.data_len = PARAGON_SEEK_FILE_DATA_LEN;
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Bytes 0...3: The offset (a 32-bit integer).
                Byte 4: The direction.
        */
        BE_SET_INT_32(command.extra_bytes.fsac.data, 0, offset);
        command.extra_bytes.fsac.data[4] = direction & 0xFF;

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                This command usually does not return a data buffer, but can
                return something if there is an error. Unfortunately I don't
                know how to interpret the error, so it's pretty useless.
        */
        if (NULL != buffer) {
                free(buffer);
        }

        free(command.extra_bytes.fsac.data);
        return result;
}

int paragon_delete_file(paragon_handle_t *handle, char *path)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle || NULL == path) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_DELETE_FILE;
        command.extra_bytes.fsac.data_len =
                PARAGON_DELETE_FILE_MIN_DATA_LEN + strlen(path);
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Bytes 0...n: The absolute path to the file.
        */
        memcpy(command.extra_bytes.fsac.data, path, strlen(path));

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                This command usually does not return a data buffer, but can
                return something if there is an error. Unfortunately I don't
                know how to interpret the error, so it's pretty useless.
        */
        if (NULL != buffer) {
                free(buffer);
        }

        free(command.extra_bytes.fsac.data);
        return result;
}

int paragon_create_directory(paragon_handle_t *handle, char *path)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle || NULL == path) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_CREATE_DIR;
        command.extra_bytes.fsac.data_len =
                PARAGON_CREATE_DIR_MIN_DATA_LEN + strlen(path);
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Bytes 0...3: Folder Attribute (0x00000010)
                Bytes 0...n: The absolute path to the file.
        */
        BE_SET_INT_32(command.extra_bytes.fsac.data, 0,
                PARAGON_FSAC_DIRECTORY);
        memcpy(command.extra_bytes.fsac.data + 4, path, strlen(path));

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                This command usually does not return a data buffer, but can
                return something if there is an error. Unfortunately I don't
                know how to interpret the error, so it's pretty useless.
        */
        if (NULL != buffer) {
                free(buffer);
        }

        free(command.extra_bytes.fsac.data);
        return result;
}

int paragon_delete_directory(paragon_handle_t *handle, char *path)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle || NULL == path) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_FSAC;
        command.extra_bytes.fsac.command = PARAGON_FSAC_DELETE_DIR;
        command.extra_bytes.fsac.data_len =
                PARAGON_DELETE_DIR_MIN_DATA_LEN + strlen(path);
        command.extra_bytes.fsac.data =
                malloc(command.extra_bytes.fsac.data_len);

        if (NULL == command.extra_bytes.fsac.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                We set the data bytes according to this structure:

                Bytes 0...3: Folder Attribute (0x00000010)
                Bytes 0...n: The absolute path to the file.
        */
        BE_SET_INT_32(command.extra_bytes.fsac.data, 0,
                PARAGON_FSAC_DIRECTORY);
        memcpy(command.extra_bytes.fsac.data + 4, path, strlen(path));

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.fsac.data);
                return result;
        }

        /*
                This command usually does not return a data buffer, but can
                return something if there is an error. Unfortunately I don't
                know how to interpret the error, so it's pretty useless.
        */
        if (NULL != buffer) {
                free(buffer);
        }

        free(command.extra_bytes.fsac.data);
        return result;
}
/* ------------------------------------------------------------------------ */
