/**
        \file fsac.h
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The header/interface for paragon commands to manipulate
        files and directories.
*/
#ifndef __PARAGON_FSAC_H__
#define __PARAGON_FSAC_H__
#include "handle.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief The max phone path length.
*/
#define PARAGON_FSAC_MAX_PATH_LEN (260)

/**
        \brief This structure is used to a file record. It only stores
        information about a file, not the file data itself.

        We use this structure for dealing with files.
*/
struct paragon_file_record_s {
        /**
                \brief The attributes of this file.
        */
        int attributes;
        /**
                \brief The owner of this file.
        */
        int owner;
        /**
                \brief The size in bytes.
        */
        int size;
        /**
                \brief A null terminated full path to a file.
        */
        char path[PARAGON_FSAC_MAX_PATH_LEN];
};

/**
        \brief A typedef to improve readability.
*/
typedef struct paragon_file_record_s paragon_file_record_t;

/**
        \defgroup fsac_commands FSAC Commands
                The list of possible File System Access Commands to issue
        to the phone.
        @{
*/
/**
        \brief This command is used to request all the volumes (analogous to
        partitions) on the phone.
*/
#define PARAGON_FSAC_GET_VOLUMES (0x0000000A)
/**
        \brief This command is used to request the number of bytes free on a
        specified volume on the phone.
*/
#define PARAGON_FSAC_GET_FREE_SPACE (0x0000000B)
/**
        \brief This command is used to request the number of files matching a
        filter on a specified volume on the phone.
*/
#define PARAGON_FSAC_GET_FILES_NUM (0x00000007)
/**
        \brief This command is used to request a file record from the phone.
*/
#define PARAGON_FSAC_GET_FILE_RECORD (0x00000008)
/**
        \brief This command is used to open a file on the phone.
*/
#define PARAGON_FSAC_OPEN_FILE (0x00000000)
/**
        \brief This command is used to close an open file on the phone.
*/
#define PARAGON_FSAC_CLOSE_FILE (0x00000004)
/**
        \brief This command is used to read an open file on the phone.
*/
#define PARAGON_FSAC_READ_FILE (0x00000001)
/**
        \brief This command is used to write to an open file on the phone.
*/
#define PARAGON_FSAC_WRITE_FILE (0x00000002)
/**
        \brief This command is used to seek in a file on the phone.
*/
#define PARAGON_FSAC_SEEK_FILE (0x00000003)
/**
        \brief This command is used to delete a file on the phone.
*/
#define PARAGON_FSAC_DELETE_FILE (0x00000005)
/**
        \brief This command is used to create a directory on the phone.
*/
#define PARAGON_FSAC_CREATE_DIR (0x0000000E)
/**
        \brief This command is used to delete a directory on the phone.
*/
#define PARAGON_FSAC_DELETE_DIR (0x0000000F)
/**
        @}
*/

/**
        \defgroup fsac_attributes File Attibutes
                The list of possible FSAC file attributes returned in a
        record.
        @{
*/
/**
        \brief The file is read-only (if unset, file is read-write)
*/
#define PARAGON_FSAC_READ_ONLY (0x01)
/**
        \brief The file is a directory.
*/
#define PARAGON_FSAC_DIRECTORY (0x10)
/**
        \brief Checks if a file has the directory attribute.
        \param fattr The file attribute to check.
*/
#define PARAGON_IS_DIR(fattr) (PARAGON_FSAC_DIRECTORY & (fattr))
/**
        @}
*/

/**
        \brief This function gets the available volumes on the phone.
        \param handle The libparagon handle to use.
        \param volumes A pointer to a character array to hold the volumes.
        \param volumes_len A pointer to an integer to hold the length of the
        volumes array.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.
        \warning Be sure to free '*volumes' when you are done.

        We allocate the array '*volumes' inside this function and write the
        volumes as single characters to this array. '*volumes' is NOT
        NULL-terminated.
*/
int paragon_get_volumes(paragon_handle_t *handle, char **volumes,
        int *volumes_len);

/**
        \brief This function gets the total number of free bytes on a volume.
        \param handle The libparagon handle to use.
        \param volume The volume to check space on.
        \param freespace A pointer to an integer to hold the number of free
        bytes.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        If there is an error, we will return PARAGON_FSAC_ERROR and not alter
        the contents of 'free'.
*/
int paragon_get_free_space(paragon_handle_t *handle, char volume,
        int *freespace);

/**
        \brief This function gets the total number of files matching the mask
        on 'volume' and sets the phone's internal file list.
        \param handle The libparagon handle to use.
        \param volume The volume to check space on.
        \param mask The filter to use. "*" selects everything.
        \param filecount A pointer to an integer to hold the number of files.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        The filter string is how we select which files to include in our
        count. Normally we will use "*" to select everything, but if, for
        example, we just wanted music files we could use "*.mp3" as the
        filter.
*/
int paragon_set_file_mask(paragon_handle_t *handle, char volume, char *mask,
        int *filecount);

/**
        \brief This function gets the next file record and stores it into
        'record'.
        \param handle The libparagon handle to use.
        \param record A pointer to hold the record.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.
        \warning You need to call paragon_set_file_mask() with the
        desired mask before this function is invoked, as the phone will
        operate on the filelist specified by that mask.

        This is an interesting function, in that it relies on the phone state
        set by the paragon_set_file_mask() function and previous calls
        to itself. This function is mostly for internal use, but I can see
        some external uses for it. We will probably abstract away the need
        for this function and the paragon_set_file_mask() function by
        wrapping them into a more complete function.
*/
int paragon_get_next_file_record(paragon_handle_t *handle,
        paragon_file_record_t *record);

/**
        \brief This function opens a file for reading/writing.
        \param handle The libparagon handle to use.
        \param path The path to the file to open; this is an absolute path
        (i.e "/a/mobile/myfile")
        \param attributes The attributes we set the file to.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.
        \warning The phone can only handle one open file at a time; the
        application must enforce this internally.

        This function is a little weird in that opening the file sets the
        file attributes; ideally an application that uses this function
        should check if the file exists and preserve the attributes.
*/
int paragon_open_file(paragon_handle_t *handle, char *path, int attributes);

/**
        \brief This function closes the currently open file on the phone.
        \param handle The libparagon handle to use.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.
        \warning The application must keep track of whether a file is open
        or not, since I don't know how to query the phone for open file
        status.

        We should close files after we're done with them; the phone can only
        handle one open file at a time.
*/
int paragon_close_file(paragon_handle_t *handle);

/**
        \brief This function reads the currently open file on the phone.
        \param handle The libparargon handle to use.
        \param bytes The number of bytes requested.
        \param recv A pointer to hold the file data.
        \param recv_size A pointer to hold the number of bytes in 'recv'.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.
        \warning The application must keep track of whether a file is open
        or not, since I don't know how to query the phone for open file
        status.

        The read file command continues reading from the current position
        in the file; we can reset this with the seek command, although many
        of the parameters are really a mystery. This function actually serves
        as a wrapper to a more basic function which can only request one
        kilobyte at a time (this function has no such restriction).
*/
int paragon_read_file(paragon_handle_t *handle, int bytes,
        unsigned char **recv, int *recv_size);

/**
        \brief This function writes to the currently open file on the phone.
        \param handle The libparargon handle to use.
        \param trans A pointer to the file data.
        \param trans_size The number of bytes in 'trans'.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.
        \warning The application must keep track of whether a file is open
        or not, since I don't know how to query the phone for open file
        status.

        The write file command continues reading from the current position
        in the file; we can reset this with the seek command, although many
        of the parameters are really a mystery. This function actually serves
        as a wrapper to a more basic function which can only request one
        kilobyte at a time (this function has no such restriction).
*/
int paragon_write_file(paragon_handle_t *handle, unsigned char *trans,
        int trans_size);

/**
        \brief This function seeks within the currently open file on the
        phone.
        \param handle The libparargon handle to use.
        \param offset The offset to seek to.
        \param direction The direction to seek from.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.
        \warning The application must keep track of whether a file is open
        or not, since I don't know how to query the phone for open file
        status.

        The seek file command is extremely poorly documented. The only
        set of parameters that I know to work are offset = 0 and
        direction = 0. This reliably seeks to the beginning of the file; any
        other combination of parameters should be tested thoroughly before
        deployment.
*/
int paragon_seek_file(paragon_handle_t *handle, int offset, int direction);

/**
        \brief This function deletes a file on the phone.
        \param handle The libparargon handle to use.
        \param path The absolute path to the file.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        The delete file command takes a single absolute path as an argument
        and deletes the file at that location, if any. It will not work on
        files with the directory attribute; use paragon_delete_directory()
        to remove directories.
*/
int paragon_delete_file(paragon_handle_t *handle, char *path);

/**
        \brief This function creates a directory on the phone.
        \param handle The libparargon handle to use.
        \param path The absolute path to the directory.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        The create directory command is interesting in that it simply makes a
        file with the correct attributes, but must be called through a
        different command.
*/
int paragon_create_directory(paragon_handle_t *handle, char *path);

/**
        \brief This function deletes a directory on the phone.
        \param handle The libparargon handle to use.
        \param path The absolute path to the directory.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        The delete directory command is used exclusively to remove
        directories; you cannot delete directories as files, enough though
        internally they appear to be similar.
*/
int paragon_delete_directory(paragon_handle_t *handle, char *path);
/* ------------------------------------------------------------------------ */
#endif
