/**
        \file handle.h
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The header for the libparagon handle.
*/
#ifndef __PARAGON_HANDLE_H__
#define __PARAGON_HANDLE_H__
#include <openusb.h>
#include <termios.h>
#include <stdarg.h>
#include "modem.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief This is the typedef for the logging function for libparagon.
*/
typedef void (paragon_logger_func_t)(int, const char *, va_list);

/**
        \brief This structure contains library specific data, such as the
        device handles, endpoint descriptors, phone state, etc.
*/
struct paragon_handle_s {
        /**
                \brief The handle to the openusb library instance.

                We use this structure in most communications with libopenusb.
        */
        openusb_handle_t openusbh;
        /**
                \brief The handle to an opened device.

                We use this in conjunction with the other openusb structures
                to communicate with the cell phone.
        */
        openusb_dev_handle_t hdev;
        /**
                \brief A list of usb devices.

                This is the list paragon_init() scans to find the appropriate
                devices.
        */
        openusb_devid_t *devs;
        /**
                \brief The number of devices in the 'devs' list.
        */
        unsigned int devnum;
        /**
                \brief The Vendor IDs we're interested in.
        */
        int vid;
        /**
                \brief The Product IDs we're interested in.
        */
        int pid;
        /**
                \brief The device index in our 'devs' list we want.
        */
        int dev_idx;
        /**
                \brief The configuration index to use.
        */
        int config_idx;
        /**
                \brief The interface index to use.
        */
        int interface_idx;
        /**
                \brief The altsetting index to use.

                Right now, this will always be 0. If libopenusb ever changes
                this, we may have different values here
        */
        int altsetting_idx;
        /**
                \brief The mode of the phone pointed to by this handle.

                We use this to determine how to get the phone to switch
                modes.
        */
        int mode;
        /**
                \brief The in endpoint of this device;
        */
        int ep_in;
        /**
                \brief The out endpoint of this device;
        */
        int ep_out;
        /**
                \brief The interface number of this device.
        */
        int ifnum;
        /**
                \brief The packet ID of the next packet on this phone.
        */
        int packet_id;
        /**
                \brief ACM device handle.

                If we are in ACM mode, this is how we communicate with the
                modem; otherwise, we ignore this value.
        */
        modem_t *acm;
        /**
                \brief The logging function to use.

                We do this so we can make custom loggers if we desire
                different debugging levels or whatnot.
        */
        paragon_logger_func_t *logger;
};

/**
        \brief This structure contains initialzation parameters for
        libparagon.

        We pass this structure to paragon_init() if we want more control over
        the intialization process.
*/
struct paragon_init_s {
        /**
                \brief The path to the ACM (modem) device.

                This is usually an absolute path (such as /dev/modem),
                although I have seen relative pathnames work fine. But you
                really should use an absolute path here. This is only
                used if the initial_mode is PARAGON_MODE_ACM.
        */
        char *acm_dev_path;
        /**
                \brief The vendor ID to check for when scanning USB devices.

                If we use PARAGON_ANY_VENDOR, we will scan devices ignoring
                the vendor id parameter.
        */
        int vendor_id;
        /**
                \brief The product ID to check for when scanning USB devices.

                If we use PARAGON_ANY_PRODUCT, we will scan devices ignoring
                the product id parameter.
        */
        int product_id;
        /**
                \brief The mode the phone is assumed to be in upon
                initialization.

                Normally it will be in either the AT or MEMCARD modes, but
                for example if we know the phone will be in P2K mode we
                can specify that here. Using PARAGON_MODE_AUTO here will
                cause paragon_init() to try to guess the phone mode.
        */
        int initial_mode;
        /**
                \brief The logging function to use.

                We do this so we can make custom loggers if we desire
                different debugging levels or whatnot.
        */
        paragon_logger_func_t *logger;
};

/**
        \brief A typedef to improve readability.
*/
typedef struct paragon_handle_s paragon_handle_t;
/**
        \brief A typedef to improve readability.
*/
typedef struct paragon_init_s paragon_init_t;

/**
        \brief Used during initialization if we don't want to limit ourselves
        only to certain vendor ids.
        \warning This depends on special libopenusb behaviour, so perhaps
        we should find a way to do without this.
*/
#define PARAGON_ANY_VENDOR (-1)
/**
        \brief Used during initialization if we don't want to limit ourselves
        only to certain product ids.
        \warning This depends on special libopenusb behaviour, so perhaps
        we should find a way to do without this.
*/
#define PARAGON_ANY_PRODUCT (-1)
/* ------------------------------------------------------------------------ */
#endif
