/**
        \file logger.c
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The implementation of the libparagon logging functions.
*/
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "logger.h"
#include "handle.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief Prints a single line hexdump from 'data' starting from
        'offset'.
        \param handle The libparagon handle to use.
        \param data The data to print out.
        \param bytes The size of 'data' in bytes.
        \param offset The offset to start from, in bytes.

        This function does some rudimentary checking and formatting and
        then sends the output to the paragon_log() function. If the offset
        is too large, this function will do nothing. If there are less bytes
        than specified in INPUT_SIZE remaining, this function will pad the
        output with spaces to keep the columns aligned. It is the core
        behind the paragon_print_hexdump() function.
*/
void paragon_print_hex_line(paragon_handle_t *handle, unsigned char *data,
        int bytes, int offset);

/**
        \brief The size of the header for the hexdump.
*/
const int HEADER_SIZE = 9;
/**
        \brief The number of bytes per line expected for the hexdump.
*/
const int INPUT_SIZE= 16;
/**
        \brief The number of padding spaces for the hexdump.
*/
const int EXTRA_PADDING_SIZE = 1;
/**
        \brief The character to represent non-printable characters.
*/
const char NON_PRINTING_CHAR = '.';

void paragon_print_hex_line(paragon_handle_t *handle, unsigned char *data,
        int bytes, int offset)
{
        int high, low;
        int i, p;
        unsigned char *line;
        int line_size;
        char *buffer;
        int buflen;

        if (NULL == handle || NULL == data) {
                return;
        }

        line_size = (bytes - offset);

        if (line_size < 1) {
                return;
        }

        if (line_size > INPUT_SIZE) {
                line_size = INPUT_SIZE;
        }

        /*
                We are formatting output like this:

                [HEADER][PADDING][HEX DATA][PADDING][ASCII DATA]

                Our header size and padding are given to us, and for every
                byte in the line (our input size) we get:

                xx[space]

                So there are 3 characters / byte in the hex data section of
                the output.

                For every byte in the line, we get:

                c

                So there is 1 character / byte in the ascii data section of
                the output.

                Combining this all and adding on character for the
                terminating NULL, we get the buffer length below.
        */
        buflen = 1 + HEADER_SIZE + EXTRA_PADDING_SIZE * 2 + INPUT_SIZE * 4;
        buffer = calloc(buflen, sizeof(char));

        if (NULL == buffer) {
                return;
        }

        line = data + offset;
        high = (offset / 0x10000) & 0xFFFF;
        low = (offset % 0x10000) & 0xFFFF;

        sprintf(buffer, "%04x%04x ", high, low);

        /*
                Print the padding between the header and the hex data.
        */
        for (i = 0, p = strlen(buffer); i < EXTRA_PADDING_SIZE; i++, p++) {
                sprintf(buffer + p, " ");
        }

        /*
                Print the hex data.
        */
        for (i = 0; i < line_size; i++, p += 3) {
                sprintf(buffer + p, "%02x ", line[i]);
        }

        /*
                If there are not enough input bytes, pad the remaining bytes
                as spaces.
        */
        for (; i < INPUT_SIZE; i++, p += 3) {
                sprintf(buffer + p, "   ");
        }

        /*
                Print the padding between the hex data and the ascii data.
        */
        for (i = 0; i < EXTRA_PADDING_SIZE; i++, p++) {
                sprintf(buffer + p, " ");
        }

        /*
                Print the ascii data.
        */
        for (i = 0; i < line_size; i++, p++) {
                if (isprint(line[i])) {
                        sprintf(buffer + p, "%c", line[i]);
                } else {
                        sprintf(buffer + p, "%c", NON_PRINTING_CHAR);
                }
        }

        /*
                Write our formatted data to the logger.
        */
        paragon_log(handle, PL_INFO, "%s\n", buffer);
        free(buffer);

        return;
}

void paragon_print_hexdump(paragon_handle_t *handle, unsigned char *data,
        int bytes)
{
        int offset = 0;
        if (NULL == data || bytes < 1) {
                return;
        }

        paragon_log(handle, PL_INFO, "\n");

        do {
                paragon_print_hex_line(handle, data, bytes, offset);
                offset += INPUT_SIZE;
        } while (offset < bytes);

        paragon_log(handle, PL_INFO, "\n");

        return;
}

void paragon_log(paragon_handle_t *handle, int level, const char *msg, ...)
{
        va_list ap;

        if (NULL == handle) {
                return;
        }

        if (NULL == handle->logger) {
                return;
        }

        va_start(ap, msg);

        /*
                Let's hope the logger knows what it's doing...
        */
        (*handle->logger)(level, msg, ap);

        va_end(ap);

        return;
}

void paragon_default_logger(int level, const char *msg, va_list ap)
{
        /*
                The default logger just prints out ALL messages to stderr.
        */
        switch (level) {
                case PL_ERROR:
                        fprintf(stderr, "(EE) ");
                break;
                case PL_WARNING:
                        fprintf(stderr, "(WW) ");
                break;
                case PL_INFO:
                        fprintf(stderr, "(II) ");
                break;
                default:
                        fprintf(stderr, "(%d) ", level);
        }
        vfprintf(stderr, msg, ap);
        return;
}
/* ------------------------------------------------------------------------ */
