/**
        \file modem.h
        \author Sachith Dunatunga
        \date 17.04.09
        \brief The header/interface for the modem interaction functions.
*/
#ifndef __PARAGON_MODEM_H__
#define __PARAGON_MODEM_H__
#include <termios.h>

/* ---modem---------------------------------------------------------------- */
/**
        \brief The structure for a modem device.
*/
struct modem_s {
        /**
                \brief This the the file descriptor for the modem device.

                This file descriptor can be used by standard UNIX file
                functions such as read or even in more advanced cases such
                as select.
        */
        int fd;
        /**
                \brief The path to the modem.
        */
        char *path;
        /**
                \brief The initial settings of the modem.
        */
        struct termios old_settings;
        /**
                \brief The current settings of the modem.
        */
        struct termios settings;
};

/**
        \brief A typedef to improve readability.
*/
typedef struct modem_s modem_t;

/**
        \brief Tries to open the modem device given.
        \param modem The full path to the modem device (i.e. "/dev/modem").
        \return NULL on error, an modem_t structure otherwise.

        This function is mostly just a wrapper to open(), although it also
        does initialization of the modem device by sending the AT command.
*/
modem_t *modem_open(char *modem);

/**
        \brief Closes the (open) modem given by fd.
        \param handle The modem to close.
        \return -1 on error, 0 on success.

        This function is mainly a wrapper to close() and free().
*/
int modem_close(modem_t *handle);

/**
        \brief Reads data from the modem.
        \param handle The modem to read from.
        \param bytes A pointer to store the number of bytes read.
        \return NULL on error, a pointer to the beginning of the data
        otherwise.

        This function is just a wrapper for ioctl() and read(). You should
        free() the returned pointer to avoid memory leaks.
*/
char *modem_read(modem_t *handle, int *bytes);

/**
        \brief Writes data to the modem
        \param handle The modem to write to.
        \param data The data to write to the modem.
        \param bytes The size of the data we are sending in bytes.
        \return -1 on error, the number of bytes written otherwise.
*/
int modem_write(modem_t *handle, char *data, int bytes);
/* ------------------------------------------------------------------------ */
#endif
