/**
        \file modes.c
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The implementation of the modeswitching functions.
*/
#include <stdlib.h>
#include <string.h>
#include <openusb.h>
#include <assert.h>
#include "error.h"
#include "handle.h"
#include "logger.h"
#include "modem.h"
#include "modes.h"
#include "scanner.h"
#include "timings.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief Resets the device handles after a modeswitch.
        \param handle The libparagon handle to use.

        This function is used to reset all now-invalid device properties in
        the libparagon handle. This happens when, for example, we switch
        modes and the phone reconnects to the bus.
*/
void paragon_reset_devhandles(paragon_handle_t *handle);
/**
        \brief Switches the phone to AT mode.
        \param handle The libparagon handle to use.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        This function will bring the phone from another mode to AT mode.
*/
int paragon_to_AT_mode(paragon_handle_t *handle);
/**
        \brief Switches the phone to Memcard mode.
        \param handle The libparagon handle to use.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        This function will bring the phone from another mode to Memcard mode.
*/
int paragon_to_MEM_mode(paragon_handle_t *handle);
/**
        \brief Switches the phone to Paragon 2000/Paragon 2005 mode.
        \param handle The libparagon handle to use.
        \return PARAGON_SUCCESS on success, an integer less than 0 on error.

        This function will bring the phone from another mode to P2K/P2K05
        mode.
*/
int paragon_to_P2K_mode(paragon_handle_t *handle);

void paragon_reset_devhandles(paragon_handle_t *handle)
{
        if (NULL != handle) {
                handle->hdev = 0;
                handle->devs = NULL;
                handle->dev_idx = 0;
                handle->config_idx = 0;
                handle->interface_idx = 0;
                handle->altsetting_idx = 0;
                handle->ep_in = -1;
                handle->ep_out = -1;
                handle->ifnum = -1;
                handle->packet_id = 0;
        }

        return;
}

int paragon_to_AT_mode(paragon_handle_t *handle)
{
        struct openusb_request_handle rqhandle;
        openusb_ctrl_request_t ctrl;
        int result = PARAGON_NO_PHONE;

        assert(NULL != handle);

        switch (handle->mode) {
                case PARAGON_MODE_AT:
                break;
                case PARAGON_MODE_ACM:
                break;
                case PARAGON_MODE_MEMCARD:
                break;
                case PARAGON_MODE_P2K:
                        /*
                                This control transfer actually doesn't
                                transfer data. Set the timeout for 5000
                                milliseconds.
                        */
                        ctrl.payload = NULL;
                        ctrl.length = 0;
                        ctrl.timeout = PARAGON_CTRL_DEFAULT_TIMEOUT;

                        /*
                                The 6th bit set means this is a
                                vendor-specific request.
                        */
                        ctrl.setup.bmRequestType = 0x40;
                        /*
                                This is the actual request. I'm assuming it
                                means 'change mode'.
                        */
                        ctrl.setup.bRequest = 0x01;
                        /*
                                This is a parameter for our request. Again I
                                don't really know what it means, but it's
                                supposed to switch to AT mode.
                        */
                        ctrl.setup.wValue = 0x00;
                        /*
                                This is yet another parameter for our request.
                        */
                        ctrl.setup.wIndex = 0x01;

                        /*
                                Setup our request handler.
                        */
                        rqhandle.dev = handle->hdev;
                        rqhandle.interface = handle->ifnum;
                        rqhandle.endpoint = handle->ep_out;
                        rqhandle.type = USB_TYPE_CONTROL;
                        rqhandle.req.ctrl = &ctrl;

                        if (OPENUSB_SUCCESS == openusb_xfer_wait(&rqhandle)) {
                                /*
                                        Sleep a bit because the modeswitch
                                        takes some time.
                                */
                                paragon_usleep(PARAGON_MODESWITCH_WAIT);
                                /*
                                        Our handles and such are now invalid,
                                        so reset them.
                                */
                                paragon_reset_devhandles(handle);
                                result = paragon_scan(handle,
                                        PARAGON_ANY_VENDOR,
                                        PARAGON_ANY_PRODUCT,
                                        PARAGON_MODE_AT);
                        } else {
                                result = PARAGON_MODE_ERROR;
                        }
                break;
                default:
                        result = PARAGON_MODE_ERROR;
        }

        return result;
}

int paragon_to_MEM_mode(paragon_handle_t *handle)
{
        int result = PARAGON_NO_PHONE;

        return result;
}

int paragon_to_P2K_mode(paragon_handle_t *handle)
{
        struct openusb_request_handle rqhandle;
        openusb_bulk_request_t bulk;
        int result = PARAGON_NO_PHONE;
        char at_init_string[] = "AT S7=45 S0=0 L1 V1 X4 &c1 E0 Q0\r";
        char at_magic_string[] = "AT+MODE=8\r";
        char acm_magic_string[] = "AT+MODE=8\r";
        /*
                This magic string actually doesn't work on my phone. Not
                too sure what's going on here.
        */
        unsigned char memcard_magic_string[] = {
                0x55, 0x53, 0x42, 0x43, 0x08, 0xC0, 0xD9, 0x85,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C, 0xD6,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00
        };
        int len;

        assert(NULL != handle);

        switch (handle->mode) {
                case PARAGON_MODE_AT:
                        /*
                                Initialize the AT device to recieve our
                                modeswitch command. The timeout is 5000
                                milliseconds.
                        */
                        bulk.payload = (unsigned char *)at_init_string;
                        bulk.timeout = PARAGON_BULK_DEFAULT_TIMEOUT;
                        bulk.length = strlen(at_init_string);

                        /*
                                Setup our request handler.
                        */
                        rqhandle.dev = handle->hdev;
                        rqhandle.interface = handle->ifnum;
                        rqhandle.endpoint = handle->ep_out;
                        rqhandle.type = USB_TYPE_BULK;
                        rqhandle.req.bulk = &bulk;

                        if (OPENUSB_SUCCESS == openusb_xfer_wait(&rqhandle)) {
                                /*
                                        Sleep a bit to wait for the
                                        initialization to complete.
                                */
                                paragon_usleep(PARAGON_AT_INIT_WAIT);
                        } else {
                                /*
                                        There's a problem, we should abort.
                                */
                                result = PARAGON_MODE_ERROR;
                                break;
                        }

                        /*
                                Tell it to transfer our magic string to the
                                AT device. The timeout is 5000 milliseconds.
                        */
                        bulk.payload = (unsigned char *)at_magic_string;
                        bulk.timeout = PARAGON_BULK_DEFAULT_TIMEOUT;
                        bulk.length = strlen(at_magic_string);

                        /*
                                Setup our request handler.
                        */
                        rqhandle.dev = handle->hdev;
                        rqhandle.interface = handle->ifnum;
                        rqhandle.endpoint = handle->ep_out;
                        rqhandle.type = USB_TYPE_BULK;
                        rqhandle.req.bulk = &bulk;

                        if (OPENUSB_SUCCESS == openusb_xfer_wait(&rqhandle)) {
                                /*
                                        Sleep a bit because the modeswitch
                                        takes some time.
                                */
                                paragon_usleep(PARAGON_MODESWITCH_WAIT);
                                /*
                                        Our handles and such are now invalid,
                                        so reset them.
                                */
                                paragon_reset_devhandles(handle);
                                result = paragon_scan(handle,
                                        PARAGON_ANY_VENDOR,
                                        PARAGON_ANY_PRODUCT,
                                        PARAGON_MODE_P2K);
                        } else {
                                result = PARAGON_MODE_ERROR;
                        }
                break;
                case PARAGON_MODE_ACM:
                        if (NULL == handle->acm) {
                                result = PARAGON_MODE_ERROR;
                                break;
                        }
                        len = strlen(acm_magic_string);
                        modem_write(handle->acm, acm_magic_string, len);
                        /*
                                Sleep a bit because the modeswitch takes some
                                time.
                        */
                        paragon_usleep(PARAGON_MODESWITCH_WAIT);
                        result = paragon_scan(handle, PARAGON_ANY_VENDOR,
                                PARAGON_ANY_PRODUCT, PARAGON_MODE_P2K);
                break;
                case PARAGON_MODE_MEMCARD:
                        /*
                                Tell it to transfer our magic string to the
                                Memcard device. The timeout is 1000
                                milliseconds.
                        */
                        bulk.payload = memcard_magic_string;
                        bulk.timeout = PARAGON_BULK_DEFAULT_TIMEOUT;
                        bulk.length = sizeof(memcard_magic_string);

                        rqhandle.dev = handle->hdev;
                        rqhandle.interface = handle->ifnum;
                        rqhandle.endpoint = handle->ep_out;
                        rqhandle.type = USB_TYPE_BULK;
                        rqhandle.req.bulk = &bulk;

                        if (OPENUSB_SUCCESS == openusb_xfer_wait(&rqhandle)) {
                                /*
                                        Sleep a bit because the modeswitch
                                        takes some time.
                                */
                                paragon_usleep(PARAGON_MODESWITCH_WAIT);
                                /*
                                        Our handles and such are now invalid,
                                        so reset them.
                                */
                                paragon_reset_devhandles(handle);
                                result = paragon_scan(handle,
                                        PARAGON_ANY_VENDOR,
                                        PARAGON_ANY_PRODUCT,
                                        PARAGON_MODE_P2K);
                        } else {
                                result = PARAGON_MODE_ERROR;
                        }
                break;
                case PARAGON_MODE_P2K:
                        result = PARAGON_SUCCESS;
                break;
                default:
                        result = PARAGON_MODE_ERROR;
        }

        return result;
}

int paragon_set_mode(paragon_handle_t *handle, int mode)
{
        int result = PARAGON_NO_PHONE;

        if (NULL == handle) {
                return PARAGON_BADARG;
        }

        switch (mode) {
                case PARAGON_MODE_AT:
                        result = paragon_to_AT_mode(handle);
                        if (result == PARAGON_SUCCESS) {
                                handle->mode = PARAGON_MODE_AT;
                        }
                break;
                case PARAGON_MODE_ACM:
                        /*
                                We can't really switch to ACM mode, so return
                                an error.
                        */
                        result = PARAGON_BADARG;
                break;
                case PARAGON_MODE_MEMCARD:
                        result = paragon_to_MEM_mode(handle);
                        if (result == PARAGON_SUCCESS) {
                                handle->mode = PARAGON_MODE_MEMCARD;
                        }
                break;
                case PARAGON_MODE_P2K:
                        result = paragon_to_P2K_mode(handle);
                        if (result == PARAGON_SUCCESS) {
                                handle->mode = PARAGON_MODE_P2K;
                        }
                break;
                default:
                        result = PARAGON_BADARG;
        }

        return result;
}

int paragon_get_mode(paragon_handle_t *handle, int *mode)
{
        int result = PARAGON_NO_PHONE;

        if (NULL == handle || NULL == mode) {
                return PARAGON_BADARG;
        }

        /*
                We assume the phone data is still valid if we still have
                valid handles to it.
        */
        if (handle->acm != NULL ||
                (0 != handle->hdev && NULL != handle->devs)) {
                *mode = handle->mode;
                result = PARAGON_SUCCESS;
        }

        return result;
}
/* ------------------------------------------------------------------------ */
