/**
        \file paragon.c
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The implementation of the libparagon functions.
*/
#include <stdlib.h>
#include "error.h"
#include "logger.h"
#include "modem.h"
#include "modes.h"
#include "scanner.h"
#include "paragon.h"

/* ---libparagon----------------------------------------------------------- */
int paragon_init(paragon_handle_t *handle, paragon_init_t *params)
{
        int result;
        int mode;

        /*
                This is exteremly bad news, we need a handle to use
                libparagon, so bail.
        */
        if (NULL == handle) {
                return PARAGON_BADARG;
        }

        /*
                Load up the default parameters if we weren't given any.
        */
        if (NULL == params) {
                handle->vid = PARAGON_ANY_VENDOR;
                handle->pid = PARAGON_ANY_PRODUCT;
                mode = PARAGON_MODE_AUTO;
                handle->logger = paragon_default_logger;
        } else {
                handle->vid = params->vendor_id;
                handle->pid = params->product_id;
                mode = params->initial_mode;
                handle->logger = params->logger;
        }

        handle->hdev = 0;
        handle->devs = NULL;
        handle->dev_idx = 0;
        handle->config_idx = 0;
        handle->interface_idx = 0;
        handle->altsetting_idx = 0;
        handle->ep_in = -1;
        handle->ep_out = -1;
        handle->ifnum = -1;
        handle->packet_id = 0;

        /*
                Decide if we need the acm device. Otherwise set it to an
                invalid handle so we don't accidentally call modem functions
                on it.
        */
        if ((PARAGON_MODE_ACM == mode) && (NULL != params->acm_dev_path)) {
                paragon_log(handle, PL_INFO, "Trying to open modem: %s\n",
                        params->acm_dev_path);
                handle->acm = modem_open(params->acm_dev_path);
                if (NULL != handle->acm) {
                        paragon_log(handle, PL_INFO, "Opened %s.\n",
                                params->acm_dev_path);
                        handle->mode = mode;
                        result = PARAGON_SUCCESS;
                } else {
                        paragon_log(handle, PL_ERROR, "Couldn't open %s.\n",
                                params->acm_dev_path);
                        handle->mode = PARAGON_MODE_UNKNOWN;
                        result = PARAGON_MODE_ERROR;
                }
        } else {
                handle->acm = NULL;
                result = PARAGON_MODE_ERROR;
        }

        /*
                The first parameter is the flags; right now they are
                reservered so we just use 0 (nothing set).
        */
        if (OPENUSB_SUCCESS != openusb_init(0, &(handle->openusbh))) {
                paragon_log(handle, PL_ERROR, "openusb_init failed.\n");
                return PARAGON_OPENUSB_ERROR;
        }

        paragon_log(handle, PL_INFO, "libopenusb initialized.\n");

        /*
                We are spoonfed device information if we're in ACM mode;
                otherwise we have to scan the bus to determine information.
        */
        if (PARAGON_MODE_ACM != mode) {
                result = paragon_scan(handle, handle->vid, handle->pid, mode);
        }

        return result;
}

int paragon_fini(paragon_handle_t *handle)
{
        int result = PARAGON_SUCCESS;

        if (NULL == handle) {
                return PARAGON_BADARG;
        }

        /*
                Close modem device...
        */
        if (NULL != handle->acm) {
                modem_close(handle->acm);
                handle->acm = NULL;
        }

        /*
                Release interfaces...
        */
        if (-1 != handle->ifnum) {
                openusb_release_interface(handle->hdev, handle->ifnum);
                handle->ifnum = -1;
        }

        /*
                Close USB devices...
        */
        if (0 != handle->hdev) {
                openusb_close_device(handle->hdev);
                handle->hdev = 0;
        }

        /*
                Free device list...
        */
        if (NULL != handle->devs) {
                openusb_free_devid_list(handle->devs);
                handle->devs = NULL;
        }

        /*
                Close libopenusb...
        */
        if (0 != handle->openusbh) {
                openusb_fini(handle->openusbh);
        }

        return result;
}
/* ------------------------------------------------------------------------ */
