/**
        \file scanner.c
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The implementation of the libparagon bus scanning functions.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openusb.h>
#include <assert.h>
#include "error.h"
#include "handle.h"
#include "logger.h"
#include "modes.h"
#include "util.h"
#include "scanner.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief Scans all devices for an interface corresponding to 'mode'.
        \param handle The libparagon handle to use.
        \param mode The mode to check for.
        \return PARAGON_SUCCESS on the interface found, an integer less than
        0 on error.

        This function is used as a helper to the higher level scanning
        function. It relies on the handle's device list already being filled.
*/
int paragon_scan_devices(paragon_handle_t *handle, int mode);
/**
        \brief Scans all configurations for an interface corresponding to
        'mode'.
        \param handle The libparagon handle to use.
        \param dev The index of the device to scan.
        \param mode The mode to check for.
        \return PARAGON_SUCCESS on the interface found, an integer less than
        0 on error.

        This function is used as a helper to the higher level scanning
        function. It relies on the handle's device list already being filled.
*/
int paragon_scan_configs(paragon_handle_t *handle, int dev, int mode);
/**
        \brief Scans all interfaces for an interface corresponding to 'mode'.
        \param handle The libparagon handle to use.
        \param d The index of the device to scan.
        \param c The index of the configuration to scan.
        \param mode The mode to check for.
        \return PARAGON_SUCCESS on the interface found, an integer less than
        0 on error.

        This function is used as a helper to the higher level scanning
        function. It relies on the handle's device list already being filled.
*/
int paragon_scan_interfaces(paragon_handle_t *handle, int d, int c, int mode);

int paragon_scan_devices(paragon_handle_t *handle, int mode)
{
        int result = PARAGON_NO_PHONE;
        int i;

        assert(NULL != handle);
        assert(NULL != handle->devs);

        for (i = 0; i < handle->devnum; i++) {
                paragon_log(handle, PL_INFO, "Device %d of %d:\n", i + 1,
                        handle->devnum);
                result = paragon_scan_configs(handle, i, mode);
                if (PARAGON_SUCCESS == result) {
                        handle->dev_idx = i;
                        result = PARAGON_SUCCESS;
                        break;
                }
        }

        return result;
}

int paragon_scan_configs(paragon_handle_t *handle, int dev, int mode)
{
        openusb_dev_data_t *data;
        int result = PARAGON_NO_PHONE;
        int configs;
        int i;

        assert(NULL != handle);
        assert(NULL != handle->devs);

        /*
                The 3rd argument is a flags setting; this is reserved so we
                just pass 0 for now.
        */
        i = openusb_get_device_data(handle->openusbh, handle->devs[dev], 0,
                &data);
        if (OPENUSB_SUCCESS != i) {
                paragon_log(handle, PL_ERROR,
                        "Couldn't get device data.\n");
                return PARAGON_OPENUSB_ERROR;
        }

        i = openusb_open_device(handle->openusbh,
                handle->devs[dev], USB_INIT_DEFAULT, &(handle->hdev));
        if (OPENUSB_SUCCESS != i) {
                paragon_log(handle, PL_ERROR,
                        "Couldn't open device.\n");
                handle->hdev = 0;
        }

        configs = data->dev_desc.bNumConfigurations;
        for (i = 0; i < configs; i++) {
                paragon_log(handle, PL_INFO, "\tConfig %d of %d:\n", i + 1,
                        configs);
                result = paragon_scan_interfaces(handle, dev, i, mode);
                if (PARAGON_SUCCESS == result) {
                        handle->config_idx = i;
                        result = PARAGON_SUCCESS;
                        break;
                }
        }

        if (i >= configs) {
                /*
                        We didn't find our device, so close the handle.
                */
                openusb_close_device(handle->hdev);
                handle->hdev = 0;
        }
        return result;
}

int paragon_scan_interfaces(paragon_handle_t *handle, int d, int c, int mode)
{
        usb_config_desc_t cdesc;
        usb_interface_desc_t ifdesc;
        usb_endpoint_desc_t epdesc;
        char at_interface[] = "Motorola Data Interface";
        char mem_interface[] = "Motorola Mass Storage Interface";
        char p2k_interface[] = "Motorola Test Command";
        char *interface_name;
        char *buf;
        int buflen;
        int result = PARAGON_NO_PHONE;
        int interfaces;
        int eps;
        int parsed;
        int i, t;

        assert(NULL != handle);
        assert(NULL != handle->devs);

        if (PARAGON_MODE_AT == mode) {
                interface_name = at_interface;
        } else if (PARAGON_MODE_MEMCARD == mode) {
                interface_name = mem_interface;
        } else if (PARAGON_MODE_P2K == mode) {
                interface_name = p2k_interface;
        } else {
                interface_name = NULL;
        }

        if (NULL == interface_name) {
                /*
                        We must have been fed a bad mode to scan for,
                        otherwise we can't get here.
                */
                paragon_log(handle, PL_ERROR, "Bad mode: Interface Name"
                        " is NULL!\n");
                return PARAGON_BADARG;
        }

        paragon_log(handle, PL_INFO, "Searching for Interface Name: %s\n",
                interface_name);

        i = openusb_parse_config_desc(handle->openusbh, handle->devs[d],
                NULL, 0, c, &cdesc);
        if (OPENUSB_SUCCESS != i) {
                paragon_log(handle, PL_ERROR, "\topenusb_parse error\n");
                return PARAGON_OPENUSB_ERROR;
        }

        /*
                Allocate space for our temporary string buffer.
        */
        buflen = strlen(interface_name) + 1;
        buf = malloc(buflen);

        if (NULL == buf) {
                return PARAGON_NO_MEMORY;
        }

        interfaces = cdesc.bNumInterfaces;
        for (i = 0; i < interfaces; i++) {
                paragon_log(handle, PL_INFO, "\tInterface %d of %d:\n", i + 1,
                        interfaces);
                parsed = openusb_parse_interface_desc(handle->openusbh,
                        handle->devs[d], NULL, 0, c, i, 0, &ifdesc);
                if (OPENUSB_SUCCESS != parsed) {
                        paragon_log(handle, PL_ERROR,
                                "Couldn't parse descriptor.\n");
                        continue;
                }
                if (0 == ifdesc.iInterface) {
                        paragon_log(handle, PL_INFO,
                                "\t\tInvalid string index.\n");
                        continue;
                }

                t = openusb_get_string_simple(handle->hdev,
                        ifdesc.iInterface, buf, buflen);
                if (t < 0) {
                        paragon_log(handle, PL_ERROR,
                                "openusb_get_string_simple failed.\n");
                        continue;
                }

                paragon_log(handle, PL_INFO, "\t\tName: %s\n", buf);

                if (0 == strcmp(interface_name, buf)) {
                        paragon_log(handle, PL_INFO,
                                "\t\tFound interface match.\n");
                        handle->interface_idx = i;
                        handle->altsetting_idx = 0;

                        handle->ep_in = -1;
                        handle->ep_out = -1;

                        eps = ifdesc.bNumEndpoints;
                        for (t = 0; t < eps; t++) {
                                paragon_log(handle, PL_INFO,
                                        "\t\tEndpoint %d of %d:\n", t + 1,
                                        eps);
                                openusb_parse_endpoint_desc(handle->openusbh,
                                        handle->devs[d], NULL, 0, c, i, 0, t,
                                        &epdesc);
                                paragon_log(handle, PL_INFO,
                                        "\t\t\tAddress: 0x%02x\n",
                                        epdesc.bEndpointAddress);
                                if (epdesc.bEndpointAddress & 0x80) {
                                        handle->ep_in =
                                                epdesc.bEndpointAddress;
                                } else {
                                        handle->ep_out =
                                                epdesc.bEndpointAddress;
                                }
                        }

                        if (handle->ifnum != -1) {
                                openusb_release_interface(handle->hdev,
                                        handle->ifnum);
                        }

                        handle->ifnum = ifdesc.bInterfaceNumber;

                        t = openusb_claim_interface(handle->hdev,
                                handle->ifnum, USB_INIT_DEFAULT);

                        if (OPENUSB_SUCCESS != t) {
                                paragon_log(handle, PL_ERROR,
                                        "Cannot claim interface.\n");
                                result = PARAGON_OPENUSB_ERROR;
                                handle->ifnum = -1;
                        } else {
                                result = PARAGON_SUCCESS;
                        }
                        break;
                }
        }

        free(buf);

        return result;
}

int paragon_scan(paragon_handle_t *handle, int vid, int pid, int mode)
{
        int result;
        int found_device;

        if (NULL == handle) {
                return PARAGON_BADARG;
        }

        if (PARAGON_MODE_AUTO == mode || PARAGON_MODE_UNKNOWN == mode) {
                /*
                        Systematically try all valid modes since we're on
                        auto or in an unstable state.
                */
                paragon_log(handle, PL_INFO, "On automatic, scanning for any"
                        " motorola devices.\n");
                result = paragon_scan(handle, vid, pid, PARAGON_MODE_AT);
                if (PARAGON_SUCCESS == result) {
                        handle->mode = PARAGON_MODE_AT;
                        return result;
                }

                result = paragon_scan(handle, vid, pid, PARAGON_MODE_MEMCARD);
                if (PARAGON_SUCCESS == result) {
                        handle->mode = PARAGON_MODE_MEMCARD;
                        return result;
                }

                result = paragon_scan(handle, vid, pid, PARAGON_MODE_P2K);
                if (PARAGON_SUCCESS == result) {
                        handle->mode = PARAGON_MODE_P2K;
                        return result;
                }

                /*
                        We couldn't find the phone, at least in any of the
                        modes we are familiar with.
                */
                return PARAGON_NO_PHONE;
        }

        if (NULL != handle->devs) {
                /*
                        We already have a a device list, so free it.
                */
                openusb_free_devid_list(handle->devs);
                handle->devs = NULL;
        }

        if (0 != handle->hdev) {
                /*
                        We already have a device open, so close it.
                */
                openusb_close_device(handle->hdev);
                handle->hdev = 0;
        }

        /*
                This is cheating a little bit, but we set PARAGON_ANY_PRODUCT
                and PARAGON_ANY_VENDOR to -1 so this call will behave as
                expected (ignores Vendor ID, Product ID if we call with
                PARAGON_ANY_*, otherwise honors the criterion). If openusb
                ever decides to change this behaviour we are hosed.
        */
        result = openusb_get_devids_by_vendor(handle->openusbh, vid, pid,
                        &(handle->devs), &(handle->devnum));

        if (OPENUSB_NULL_LIST == result) {
                /*
                        No devices matching the criterion exist.
                */
                paragon_log(handle, PL_ERROR, "Couldn't find devices.\n");
                return PARAGON_NO_PHONE;
        } else if (OPENUSB_SUCCESS != result) {
                /*
                        We have another openusb error here; get out since our
                        devids list is invalid.
                */
                paragon_log(handle, PL_ERROR, "Couldn't get devid list.\n");
                return PARAGON_OPENUSB_ERROR;
        }

        /*
                Iterate through our returned devices.
        */
        found_device = paragon_scan_devices(handle, mode);

        if (PARAGON_SUCCESS == found_device) {
                result = PARAGON_SUCCESS;
        } else {
                /*
                        Make sure we free the device list and prepare it for
                        another scan
                */
                openusb_free_devid_list(handle->devs);
                handle->devs = NULL;
                handle->hdev = 0;
                result = PARAGON_NO_PHONE;
        }

        return result;
}
/* ------------------------------------------------------------------------ */
