/**
        \file seem.c
        \author Sachith Dunatunga
        \date 21.04.09
        \brief The implementation of the libparagon commands to manipulate
        seems (Binary motorola phone settings).
*/
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "command.h"
#include "error.h"
#include "logger.h"
#include "handle.h"
#include "seem.h"

/* ---libparagon----------------------------------------------------------- */
/**
        \brief The number of bytes needed in the data section of the seem
        read command. Think of them as the argument bytes.
*/
const int PARAGON_READ_SEEM_DATA_LEN = 8;
/**
        \brief The minimum number of bytes needed in the data section of the
        seem write command. Think of them as the argument bytes.
*/
const int PARAGON_WRITE_SEEM_MIN_DATA_LEN = 8;

int paragon_read_seem(paragon_handle_t *handle, int number, int record,
        paragon_seem_t *recv)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer;
        int buflen;

        if (NULL == handle || NULL == recv) {
                return PARAGON_BADARG;
        }

        if (NULL != recv->data) {
                paragon_log(handle, PL_WARNING, "Seem data is not NULL"
                        " (possible lost pointer).\n");
        }

        recv->number = number;
        recv->record = record;
        recv->offset = 0;
        recv->data_len = 0;

        command.command = PARAGON_COMMAND_READ_SEEM;
        command.extra_len = PARAGON_READ_SEEM_DATA_LEN;
        command.extra_bytes.data = malloc(command.extra_len);

        if (NULL == command.extra_bytes.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                Setup the seem reading packet's data; it is laid out as
                follows:

                Bytes 0, 1: seem number
                Bytes 2, 3: seem record
                Bytes 4, 5: seem offset
                Bytes 6, 7: seem length

                By setting offset and length to 0, we get the entire packet,
                from the beginning. All the integers are big endian.
        */
        BE_SET_INT_16(command.extra_bytes.data, 0, recv->number);
        BE_SET_INT_16(command.extra_bytes.data, 2, recv->record);
        BE_SET_INT_16(command.extra_bytes.data, 4, recv->offset);
        BE_SET_INT_16(command.extra_bytes.data, 6, recv->data_len);

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.data);
                return result;
        }

        /*
                The data packet returned contains a zero byte at the
                beginning; it is not part of the seem data so we should
                only return the actual seem data.
        */
        recv->data_len = buflen - 1;
        recv->data = malloc(recv->data_len);
        if (NULL == recv->data) {
                free(buffer);
                free(command.extra_bytes.data);
                recv->data_len = 0;
                return PARAGON_NO_MEMORY;
        }

        memcpy(recv->data, buffer + 1, recv->data_len);
        free(buffer);
        free(command.extra_bytes.data);
        result = PARAGON_SUCCESS;

        return result;
}

int paragon_write_seem(paragon_handle_t *handle, paragon_seem_t *trans)
{
        int result = PARAGON_OPENUSB_ERROR;
        paragon_command_t command;
        unsigned char *buffer = NULL;
        int buflen;

        if (NULL == handle || NULL == trans) {
                return PARAGON_BADARG;
        }

        if (NULL == trans->data || trans->data_len < 1) {
                return PARAGON_BADARG;
        }

        command.command = PARAGON_COMMAND_WRITE_SEEM;
        command.extra_len = PARAGON_WRITE_SEEM_MIN_DATA_LEN + trans->data_len;
        command.extra_bytes.data = malloc(command.extra_len);

        if (NULL == command.extra_bytes.data) {
                return PARAGON_NO_MEMORY;
        }

        /*
                Setup the seem writing packet's data; it is laid out as
                follows:

                Bytes 0, 1: seem number
                Bytes 2, 3: seem record
                Bytes 4, 5: seem offset
                Bytes 6, 7: seem length
                Bytes 8...n: seem data

                We will assume the caller knows what they are doing with
                respect to the write. All the integers are big endian, but
                this is transparent to the caller.
        */
        BE_SET_INT_16(command.extra_bytes.data, 0, trans->number);
        BE_SET_INT_16(command.extra_bytes.data, 2, trans->record);
        BE_SET_INT_16(command.extra_bytes.data, 4, trans->offset);
        BE_SET_INT_16(command.extra_bytes.data, 6, trans->data_len);
        memcpy(command.extra_bytes.data + 8, trans->data, trans->data_len);

        result = paragon_send_command(handle, &command, &buffer, &buflen);

        if (PARAGON_SUCCESS != result) {
                free(command.extra_bytes.data);
                return result;
        }

        /*
                We don't care about the returned buffer in this case. However,
                it is usually a single zero byte:

                Byte 0: zero padding (0)
        */

        free(buffer);
        free(command.extra_bytes.data);

        return result;
}
/* ------------------------------------------------------------------------ */
