/**
        \file util.c
        \author Sachith Dunatunga
        \date 05.04.09
        \brief The implementation of utility functions for dealing with
        libopenusb.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openusb.h>
#include <errno.h>
#include "util.h"

/* ---openusb-util--------------------------------------------------------- */
/**
        \brief Converts a utf8 string to its ascii equivalent.
        \param ascii The string buffer to hold the ascii version.
        \param utf8 The string buffer that holds the utf8 version.
        \param asciibytes The size in bytes of the ascii buffer.
        \param utf8bytes The size in bytes of the utf8 buffer.
        \return The number of bytes copied into the ascii buffer.
        \warning This is for internal implementation only.

        This function simply converts from libopenusb's version of utf8 to
        a normal, ascii c character array. This is only useful for printing
        out data.
*/
int utf8_to_ascii(char *ascii, char *utf8, int asciibytes, int utf8bytes);

int utf8_to_ascii(char *ascii, char *utf8, int asciibytes, int utf8bytes)
{
        int dest = 0;
        int src = 0;

        if (NULL == utf8 || NULL == ascii) {
                return 0;
        }

        /*
                We get every even addressed byte (the lower byte in a word),
                and since the first 128 utf-8 bytes are exactly the same as
                the normal ascii bytes, we can simply copy them to the
                destination string.
        */
        while (src + 1 < utf8bytes && dest < asciibytes) {
                if (utf8[src + 1] != 0) {
                        ascii[dest] = '?';
                } else {
                        ascii[dest] = utf8[src];
                }
                dest++;
                src += 2;
        }
        ascii[dest] = '\0';

        return dest;
}

char *convert_string(usb_string_desc_t *st, int *bytes)
{
        char *convert;
        int utf8bytes;
        int asciibytes;

        if (NULL == st || NULL == bytes) {
                return NULL;
        }

        /*
                The bLength attribute is the size of the entire descriptor,
                and there are 2 bytes of header information.
        */
        utf8bytes = st->bLength - 2;
        /*
                Since this encoding of UTF-8 uses 2 bytes, we need exactly
                half as many bytes for the ascii version.
        */
        asciibytes = utf8bytes / 2;

        /*
                We need to allocate an extra byte for the terminating NULL.
        */
        *bytes = asciibytes + 1;
        convert = malloc(*bytes);
        utf8_to_ascii(convert, (char *)st->bString, asciibytes, utf8bytes);
        convert[asciibytes] = '\0';

        return convert;
}

void print_device_data(openusb_dev_data_t *data)
{
        char *manufacturer;
        char *product;
        char *serial;
        int bytes;

        if (NULL != data) {
                /*
                        We know these two will always exist so we should
                        print them at least.
                */
                printf("\t\tVendor ID: %04x\n", data->dev_desc.idVendor);
                printf("\t\tProduct ID: %04x\n", data->dev_desc.idProduct);

                /*
                        These we're not so sure about. Print them if they
                        exist.
                */
                manufacturer = convert_string(data->manufacturer, &bytes);
                product = convert_string(data->product, &bytes);
                serial = convert_string(data->serialnumber, &bytes);
                if (NULL != manufacturer) {
                        printf("\t\tManufacturer: %s\n", manufacturer);
                        free(manufacturer);
                }
                if (NULL != product) {
                        printf("\t\tProduct: %s\n", product);
                        free(product);
                }
                if (NULL != serial) {
                        printf("\t\tSerial Number: %s\n", serial);
                        free(serial);
                }
        } else {
                /*
                        This happens on my machine since USB devices can only
                        be directly accessed by root.
                */
                printf("\t\tCan't get data (probably permission error).\n");
        }

        return;
}

int openusb_get_string(openusb_dev_handle_t dev, int index, int langid,
        char *buf, size_t buflen)
{
        openusb_ctrl_request_t ctrl;

        /* if index == 0, then the caller wants to get STRING descript Zero
        * for LANGIDs
        */
        if ((buf == NULL) || (buflen == 0)) {
                return OPENUSB_BADARG;
        }

        memset(&ctrl, 0, sizeof(ctrl));
        ctrl.setup.bmRequestType = USB_REQ_DEV_TO_HOST;
        ctrl.setup.bRequest = USB_REQ_GET_DESCRIPTOR;
        ctrl.setup.wValue = (USB_DESC_TYPE_STRING << 8) + index;
        ctrl.setup.wIndex = langid;
        ctrl.payload = (uint8_t *)buf;
        ctrl.length = buflen;
        ctrl.timeout = 100;

        if (openusb_ctrl_xfer(dev, 0, 0, &ctrl) == 0) {
                return ctrl.result.transferred_bytes;
        } else {
                return -1;
        }
}

int openusb_get_string_simple(openusb_dev_handle_t dev, int index, char *buf,
        size_t buflen)
{
        char tbuf[256];
        int ret, langid, si, di;

        if (index == 0) {
                return (OPENUSB_BADARG);
        }

        if ((buf == NULL) || (buflen == 0)) {
                return (OPENUSB_BADARG);
        }

        (void) memset(buf, 0, buflen);

        /*
        * Asking for the zero'th index is special - it returns a string
        * descriptor that contains all the language IDs supported by the
        * device. Typically there aren't many - often only one. The
        * language IDs are 16 bit numbers, and they start at the third byte
        * in the descriptor. See USB 2.0 specification, section 9.6.7, for
        * more information on this.
        */
        ret = openusb_get_string(dev, 0, 0, tbuf, sizeof (tbuf));

        if (ret < 4) {
                langid = 0x409;
        } else {
                langid = tbuf[2] | (tbuf[3] << 8);
        }

        ret = openusb_get_string(dev, index, langid, tbuf, sizeof (tbuf));

        if (ret < 0) {

                return (ret);
        }
        if (tbuf[1] != USB_DESC_TYPE_STRING) {

                return (-EIO);
        }
        if (tbuf[0] > ret) {

                return (-EFBIG);
        }

        for (di = 0, si = 2; si < tbuf[0]; si += 2) {
                if (di >= (buflen - 1)) {
                        break;
                }

                if (tbuf[si + 1]) {  /* high byte */
                        buf[di++] = '?';
                } else {
                        buf[di++] = tbuf[si];
                }
        }

        buf[di] = 0;

        return (di);
}
/* ------------------------------------------------------------------------ */
