/**
        \file util.h
        \author Sachith Dunatunga
        \date 05.04.09
        \brief The header/interface for utility functions for dealing with
        libopenusb.
*/
#ifndef __OPENUSB_UTIL_H__
#define __OPENUSB_UTIL_H__
#include <openusb.h>

/* ---openusb-util--------------------------------------------------------- */
/**
        \brief Converts a string from the usb_string_desc_t type to a regular
        character array.
        \param st The usb_string_desc_t string to convert.
        \param bytes A pointer to an integer to hold the number of bytes
        in the resulting string.
        \return NULL on error, otherwise a pointer to the string.
        \warning Don't forget to free this string manually, since you have
        to keep track of it yourself.

        This function is useful to convert the utf8 in st->bString to good
        old 7-bit ascii. It is useful for feeding libopenusb strings to the
        standard c string functions.
*/
char *convert_string(usb_string_desc_t *st, int *bytes);

/**
        \brief Prints out device data for a specified device.
        \param data The data to print out.

        This is just a wrapper to printf and the convert string function
        that will print out some relevant device data.
*/
void print_device_data(openusb_dev_data_t *data);

/**
        \brief This function is used to get a string from the device pointed
        to by handle.
        \param dev The device handle to look up strings for.
        \param index The index of the string to retrieve.
        \param langid The language ID we use when getting the string (used to
        select different languages).
        \param buf The buffer to store the retrieved string.
        \param buflen The maximum size of 'buf'.
        \return The number of bytes written to 'buf' on success, less than 0
        otherwise.

        This function is actually present in the source code of openusb. Why
        its functionality isn't exposed to developers is beyond me, but it's
        quite useful.
*/
int openusb_get_string(openusb_dev_handle_t dev, int index, int langid,
        char *buf, size_t buflen);
/**
        \brief This function is used to get a string from the device pointed
        to by handle.
        \param dev The device handle to look up strings for.
        \param index The index of the string to retrieve.
        \param buf The buffer to store the retrieved string.
        \param buflen The maximum size of 'buf'.
        \return The number of bytes written to 'buf' on success, less than 0
        otherwise.

        This function is also present in the source code of openusb. It is
        different from openusb_get_string() in that it will simply take the
        first language ID it finds in the string descriptor 0 (a descriptor
        containing all the supported languages) and call openusb_get_string()
        with that language ID as a parameter. As a result it has all the
        quirks of the openusb_get_string() function.
*/
int openusb_get_string_simple(openusb_dev_handle_t dev, int index, char *buf,
        size_t buflen);
/* ------------------------------------------------------------------------ */
#endif
